from __future__ import annotations

import argparse
from collections.abc import Sequence
from pathlib import Path
from typing import Any, cast

import attrs
from typing_extensions import override

from snakebids import bidsapp
from snakebids.plugins.base import PluginBase
from snakebids.types import OptionalFilter
from snakebids.utils.utils import text_fold


class FilterParseError(Exception):
    """Raised when a magic CLI filter cannot be parsed."""

    def __init__(self, filter: str, msg: str):
        super().__init__(
            "The following filter provided by the CLI could not be parsed: "
            f"'{filter}'.\n\t{msg}".expandtabs(4)
        )

    @classmethod
    def invalid_spec(cls, filter: str, spec: str):
        """Raise if spec not recognized."""
        return cls(
            filter,
            f"':{spec}' is not a valid filter method. Must be one of 'required', "
            "'optional', 'none', 'match', or 'search'.",
        )

    @classmethod
    def missing_value(cls, filter: str, key: str, spec: str):
        """Raise if no value provided."""
        return cls(
            filter, f"':{spec}' requires a value, specified as '{key}:{spec}=VALUE'."
        )

    @classmethod
    def only_key(cls, filter: str, key: str):
        """Raise if only a key provided."""
        return cls(filter, "Filters must be specified as ENTITY[:METHOD]=VALUE.")

    @classmethod
    def unneeded_value(cls, filter: str, key: str, spec: str, value: str):
        """Raise if value provided to a boolean filter method."""
        return cls(
            filter, f"'{key}:{spec}' should not be given a value (got '={value}')"
        )


class FilterParse(argparse.Action):
    """Class for parsing CLI filters in argparse."""

    # Constructor calling
    @override
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ):
        result = {}
        if not values:
            return

        boolean_filters = {
            "optional": OptionalFilter,
            "required": True,
            "any": True,
            "none": False,
        }
        for pair in values:
            if "=" in pair:
                # split it into key and value
                key, value = cast("tuple[str, str]", pair.split("=", 1))
            else:
                key = pair
                value = None
            if ":" in key:
                key, spec = cast("tuple[str, str]", key.split(":", 1))
                spec = spec.lower()
                if spec in boolean_filters:
                    if value is not None:
                        raise FilterParseError.unneeded_value(pair, key, spec, value)
                    value = boolean_filters[spec]
                elif spec in {"match", "search"}:
                    if value is None:
                        raise FilterParseError.missing_value(pair, key, spec)
                    value = {spec: value}
                else:
                    raise FilterParseError.invalid_spec(pair, spec)
            if value is None:
                raise FilterParseError.only_key(pair, key)

            # assign into dictionary
            result[key] = value

        setattr(namespace, self.dest, result)


@attrs.define
class ComponentEdit(PluginBase):
    """Use CLI arguments to edit the filters, wildcards, and paths of components.

    Arguments are added based on the components specified in ``config``. For each
    component, a ``--filter-<comp_name>``, ``--wildcards-<comp_name>``, and a
    ``--path-<comp_name>`` argument will be added to the CLI. After parsing, these
    arguments are read and used to update the original component specification within
    config.

    Filters are specified on the CLI using ``ENTITY[:METHOD][=VALUE]``, as follows:

    1. ``ENTITY=VALUE`` selects paths based on an exact value match.
    2. ``ENTITY:match=REGEX`` and ``ENTITY:search=REGEX`` selects paths using regex
       with :func:`re.match` and :func:`re.search` respectively. This syntax can be used
       to select multiple values (e.g. ``'session:match=01|02'``).
    3. ``ENTITY:required`` selects all paths with the entity, regardless of value.
    4. ``ENTITY:none`` selects all paths without the entity.
    5. ``ENTITY:any`` removes filters for the entity.

    CLI arguments created by this plugin cannot be overridden.

    Parameters
    ----------
    components_key
        Key of component specification within the config dictionary.
    """

    components_key: str = "pybids_inputs"

    PREFIX = "plugins.componentedit"

    def __eq__(self, other: Any):
        return isinstance(other, self.__class__)

    @bidsapp.hookimpl
    def add_cli_arguments(
        self, parser: argparse.ArgumentParser, config: dict[str, Any]
    ):
        """Add filter, wildcard, and path override arguments for each component."""
        if (pybids_inputs := config.get(self.components_key)) is None:
            return

        # general parser for
        # --filter_{input_type} {key1}={value1} {key2}={value2}...
        # create filter parsers, one for each input_type
        filter_opts = parser.add_argument_group(
            "BIDS FILTERS",
            text_fold(
                """
                Update filters for input components. Each filter can be specified as a
                ENTITY=VALUE pair to select an value directly. To use regex filtering,
                ENTITY:match=REGEX or ENTITY:search=REGEX can be used for re.match() or
                re.search() respectively. Regex can also be used to select multiple
                values, e.g. 'session:match=01|02'. ENTITY:required and ENTITY:none can
                be used to require or prohibit presence of an entity in selected paths,
                respectively. ENTITY:optional can be used to remove a filter.
                """
            ),
        )

        for input_type in pybids_inputs:
            argnames = (f"--filter-{input_type}", f"--filter_{input_type}")
            filters = pybids_inputs[input_type].get("filters", {})
            arglist_default = [f"{key}={value}" for (key, value) in filters.items()]

            self.add_argument(
                filter_opts,
                *argnames,
                nargs="+",
                action=FilterParse,
                dest=f"{self.PREFIX}.filter.{input_type}",
                metavar="ENTITY[:METHOD][=VALUE]",
                help=f"(default: {' '.join(arglist_default)})",
            )

        # general parser for
        # --wildcards_{input_type} {wildcard1} {wildcard2} ...
        # create wildcards parsers, one for each input_type
        wildcards_opts = parser.add_argument_group(
            "INPUT WILDCARDS",
            "Provide entities to be used as wildcards.",
        )

        for input_type in pybids_inputs:
            argnames = (f"--wildcards-{input_type}", f"--wildcards_{input_type}")
            arglist_default = [
                f"{wc}" for wc in pybids_inputs[input_type].get("wildcards", [])
            ]

            self.add_argument(
                wildcards_opts,
                *argnames,
                nargs="+",
                dest=f"{self.PREFIX}.wildcards.{input_type}",
                metavar="WILDCARD",
                help=f"(default: {' '.join(arglist_default)})",
            )

        override_opts = parser.add_argument_group(
            "PATH OVERRIDE",
            (
                "Options for overriding BIDS by specifying absolute paths "
                "that include wildcards, e.g.: "
                "/path/to/my_data/{subject}/t1.nii.gz"
            ),
        )

        # create path override parser
        for input_type in pybids_inputs:
            argnames = (f"--path-{input_type}", f"--path_{input_type}")
            self.add_argument(
                override_opts,
                *argnames,
                default=None,
                metavar="PATH",
                dest=f"{self.PREFIX}.path.{input_type}",
            )

    @bidsapp.hookimpl
    def update_cli_namespace(self, namespace: dict[str, Any], config: dict[str, Any]):
        """Apply provided overrides to the component configuration."""
        # argparse adds filter_{input_type}
        # we want to update the pybids_inputs dict with this, then remove the
        # filter_{input_type} dict
        if (pybids_inputs := config.get(self.components_key)) is None:
            return
        for input_type in pybids_inputs:
            arg_filter_dict = self.pop(namespace, f"filter.{input_type}", None)
            if arg_filter_dict is not None:
                pybids_inputs[input_type].setdefault("filters", {})
                for entity, filter_ in arg_filter_dict.items():
                    if filter_ is OptionalFilter:
                        pybids_inputs[input_type]["filters"].pop(entity, None)
                    else:
                        pybids_inputs[input_type]["filters"][entity] = filter_

        # add cmdline defined wildcards from the list:
        # wildcards_{input_type}
        for input_type in pybids_inputs:
            wildcards_list = self.pop(namespace, f"wildcards.{input_type}", None)
            if wildcards_list is not None:
                pybids_inputs[input_type].setdefault("wildcards", [])
                pybids_inputs[input_type]["wildcards"] += wildcards_list

        # add custom input paths to
        # config['pybids_inputs'][input_type]['custom_path']
        for input_type in pybids_inputs:
            custom_path = self.pop(namespace, f"path.{input_type}", None)
            if custom_path is not None:
                pybids_inputs[input_type]["custom_path"] = Path(custom_path).resolve()



"""
This type stub file was generated by pyright.
"""

"""
Contains helper functions that involve writing operations.
"""
__all__ = ["build_path", "write_to_file"]
_PATTERN_FIND = ...

def build_path(entities, path_patterns, strict=...):  # -> list[Unknown] | None:
    """
    Constructs a path given a set of entities and a list of potential
    filename patterns to use.

    Parameters
    ----------
    entities : :obj:`dict`
        A dictionary mapping entity names to entity values.
        Entities with ``None`` or empty-string value will be removed.
        Otherwise, entities will be cast to string values, therefore
        if any format is expected (e.g., zero-padded integers), the
        value should be formatted.
    path_patterns : :obj:`str` or :obj:`list`
        One or more filename patterns to write
        the file to. Entities should be represented by the name
        surrounded by curly braces. Optional portions of the patterns
        should be denoted by square brackets. Entities that require a
        specific value for the pattern to match can pass them inside
        angle brackets. Default values can be assigned by specifying a string after
        the pipe operator. E.g., (e.g., {type<image>|bold} would only match
        the pattern if the entity 'type' was passed and its value is
        "image", otherwise the default value "bold" will be used).
    strict : :obj:`bool`
        If True, all passed entities must be matched inside a
        pattern in order to be a valid match. If False, extra entities will
        be ignored so long as all mandatory entities are found.

    Returns
    -------
    A constructed path for this file based on the provided patterns, or
    ``None`` if no path was built given the combination of entities and patterns.

    Examples
    --------
    >>> entities = {
    ...     'extension': '.nii',
    ...     'space': 'MNI',
    ...     'subject': '001',
    ...     'suffix': 'inplaneT2',
    ... }
    >>> patterns = ['sub-{subject}[/ses-{session}]/anat/sub-{subject}[_ses-{session}]'
    ...             '[_acq-{acquisition}][_ce-{ceagent}][_rec-{reconstruction}]_'
    ...             '{suffix<T[12]w|T1rho|T[12]map|T2star|FLAIR|FLASH|PDmap|PD|PDT2|'
    ...             'inplaneT[12]|angio>}{extension<.nii|.nii.gz|.json>|.nii.gz}',
    ...             'sub-{subject}[/ses-{session}]/anat/sub-{subject}[_ses-{session}]'
    ...             '[_acq-{acquisition}][_ce-{ceagent}][_rec-{reconstruction}]'
    ...             '[_space-{space}][_desc-{desc}]_{suffix<T1w|T2w|T1rho|T1map|T2map|'
    ...             'T2star|FLAIR|FLASH|PDmap|PD|PDT2|inplaneT[12]|angio>}'
    ...             '{extension<.nii|.nii.gz|.json>|.nii.gz}']
    >>> build_path(entities, patterns)
    'sub-001/anat/sub-001_inplaneT2.nii'

    >>> build_path(entities, patterns, strict=True)
    'sub-001/anat/sub-001_space-MNI_inplaneT2.nii'

    >>> entities['space'] = None
    >>> build_path(entities, patterns, strict=True)
    'sub-001/anat/sub-001_inplaneT2.nii'

    >>> # If some entity is set to None, they are dropped
    >>> entities['extension'] = None
    >>> build_path(entities, patterns, strict=True)
    'sub-001/anat/sub-001_inplaneT2.nii.gz'

    >>> # If some entity is set to empty-string, they are dropped
    >>> entities['extension'] = ''
    >>> build_path(entities, patterns, strict=True)
    'sub-001/anat/sub-001_inplaneT2.nii.gz'

    >>> # If some selector is not in the pattern, skip it...
    >>> entities['datatype'] = 'anat'
    >>> build_path(entities, patterns)
    'sub-001/anat/sub-001_inplaneT2.nii.gz'

    >>> # ... unless the pattern should be strictly matched
    >>> entities['datatype'] = 'anat'
    >>> build_path(entities, patterns, strict=True) is None
    True

    >>> # If the value of an entity is not valid, do not match the pattern
    >>> entities['suffix'] = 'bold'
    >>> build_path(entities, patterns) is None
    True

    >>> entities = {
    ...     'extension': '.bvec',
    ...     'subject': '001',
    ... }
    >>> patterns = (
    ...     "sub-{subject}[/ses-{session}]/{datatype|dwi}/sub-{subject}[_ses-{session}]"
    ...     "[_acq-{acquisition}]_{suffix|dwi}{extension<.bval|.bvec|.json|.nii.gz|.nii>|.nii.gz}"
    ... )
    >>> build_path(entities, patterns, strict=True)
    'sub-001/dwi/sub-001_dwi.bvec'

    >>> # Lists of entities are expanded
    >>> entities = {
    ...     'extension': '.bvec',
    ...     'subject': ['%02d' % i for i in range(1, 4)],
    ... }
    >>> build_path(entities, patterns, strict=True)
    ['sub-01/dwi/sub-01_dwi.bvec', 'sub-02/dwi/sub-02_dwi.bvec', 'sub-03/dwi/sub-03_dwi.bvec']

    """
    ...

def write_to_file(
    path,
    contents=...,
    link_to=...,
    copy_from=...,
    content_mode=...,
    root=...,
    conflicts=...,
):  # -> None:
    """
    Writes provided contents to a new path, or copies from an old path.

    Parameters
    ----------
    path : str
        Destination path of the desired contents.
    contents : str
        Raw text or binary encoded string of contents to write
        to the new path.
    link_to : str
        Optional path with which to create a symbolic link to.
        Used as an alternative to, and takes priority over, the contents
        argument.
    copy_from : str
        Optional filename to copy to new location. Used an alternative to, and
        takes priority over, the contents argument.
    content_mode : {'text', 'binary'}
        Either 'text' or 'binary' to indicate the writing
        mode for the new file. Only relevant if contents is provided.
    root : str
        Optional root directory that all patterns are relative
        to. Defaults to current working directory.
    conflicts : {'fail', 'skip', 'overwrite', 'append'}
        One of 'fail', 'skip', 'overwrite', or 'append'
        that defines the desired action when the output path already
        exists. 'fail' raises an exception; 'skip' does nothing;
        'overwrite' overwrites the existing file; 'append' adds a suffix
        to each file copy, starting with 1. Default is 'fail'.
    """
    ...



"""
This type stub file was generated by pyright.
"""

if False: ...

class Node:
    __slots__ = ...
    def __init__(
        self,
        tag: Any,
        value: Any,
        start_mark: Any,
        end_mark: Any,
        comment: Any = ...,
        anchor: Any = ...,
    ) -> None: ...
    @property
    def tag(self) -> Optional[str]: ...
    @tag.setter
    def tag(self, val: Any) -> None: ...
    def __repr__(self) -> Any: ...
    def dump(self, indent: int = ...) -> None: ...

class ScalarNode(Node):
    """
    styles:
      ? -> set() ? key, no value
      - -> suppressable null value in set
      " -> double quoted
      ' -> single quoted
      | -> literal style
      > -> folding style
    """

    __slots__ = ...
    id = ...
    def __init__(
        self,
        tag: Any,
        value: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        style: Any = ...,
        comment: Any = ...,
        anchor: Any = ...,
    ) -> None: ...

class CollectionNode(Node):
    __slots__ = ...
    def __init__(
        self,
        tag: Any,
        value: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        flow_style: Any = ...,
        comment: Any = ...,
        anchor: Any = ...,
    ) -> None: ...

class SequenceNode(CollectionNode):
    __slots__ = ...
    id = ...

class MappingNode(CollectionNode):
    __slots__ = ...
    id = ...
    def __init__(
        self,
        tag: Any,
        value: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        flow_style: Any = ...,
        comment: Any = ...,
        anchor: Any = ...,
    ) -> None: ...



"""Internal module for console introspection.

Module modified from [pandas](https://github.com/pandas-dev/pandas)
`pandas.io.formats.console` under BSD 3-Clause License

LICENSE
=======

BSD 3-Clause License

Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData \
Development Team
All rights reserved.

Copyright (c) 2011-2023, Open source contributors.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

import functools as ft
import sys
from shutil import get_terminal_size


def get_console_size() -> tuple[int | None, int | None]:
    """Return console size as tuple = (width, height).

    Returns (None,None) in non-interactive session.
    """
    # Consider
    # interactive shell terminal, can detect term size
    # interactive non-shell terminal (ipnb/ipqtconsole), cannot detect term
    # size non-interactive script, should disregard term size

    # in addition
    # width,height have default values, but setting to 'None' signals
    # should use Auto-Detection, But only in interactive shell-terminal.
    # Simple. yeah.

    if in_interactive_session():
        if in_ipython_frontend():
            # sane defaults for interactive non-shell terminal
            # match default for width,height in config_init

            # Hard code the default values found in pandas config
            terminal_width = 80
            terminal_height = 60
        else:
            # pure terminal
            terminal_width, terminal_height = get_terminal_size()
    else:
        terminal_width, terminal_height = None, None

    # Note if the User sets width/Height to None (auto-detection)
    # and we're in a script (non-inter), this will return (None,None)
    # caller needs to deal.
    return terminal_width, terminal_height


# ----------------------------------------------------------------------
# Detect our environment


@ft.lru_cache
def in_interactive_session() -> bool:
    """Check if we're running in an interactive shell.

    Returns
    -------
    bool
        True if running under python/ipython interactive shell.
    """

    def check_main() -> bool:
        return hasattr(sys, "ps1")

    try:
        # error: Name '__IPYTHON__' is not defined
        return __IPYTHON__ or check_main()  # type: ignore[name-defined]
    except NameError:
        return check_main()


def in_ipython_frontend() -> bool:
    """Check if we're inside an IPython zmq frontend.

    Returns
    -------
    bool
    """
    try:
        # error: Name 'get_ipython' is not defined
        ipy = get_ipython()  # type: ignore[name-defined]
        return "zmq" in str(type(ipy)).lower()  # type: ignore
    except NameError:
        pass

    return False



from __future__ import annotations

import textwrap
import warnings

from colorama import Fore, Style

WARN_TEMPLATE = f"""\
{Fore.YELLOW}{Style.BRIGHT}[{{category}}]{Style.NORMAL} {{filename}}:{{lineno}} \
{Fore.RESET}{{line}}
{{message}}

"""


def formatwarning(
    message: Warning | str,
    category: type[Warning],
    filename: str,
    lineno: int,
    line: str | None,
):
    """Format warning messages."""
    if line is None:
        try:
            import linecache  # noqa: PLC0415

            line = linecache.getline(filename, lineno).rstrip()
        except Exception:  # pragma: no cover # noqa: BLE001
            # When a warning is logged during Python shutdown, linecache
            # and the import machinery don't work anymore
            line = None
            linecache = None

    return WARN_TEMPLATE.format(
        message=textwrap.indent(
            message.args[0] if isinstance(message, Warning) else message,
            prefix="  ",
        ),
        category=category.__name__,
        filename=filename,
        lineno=lineno,
        line=line,
    )


warnings.formatwarning = formatwarning



"""
This type stub file was generated by pyright.
"""

from pathlib import Path
from types import TracebackType
from typing import Any, Optional, Union

from _ruamel_yaml import CEmitter
from ruamel.yaml.compat import StreamTextType, StreamType, VersionType, with_metaclass
from ruamel.yaml.constructor import Constructor
from ruamel.yaml.dumper import *
from ruamel.yaml.events import *
from ruamel.yaml.nodes import *
from ruamel.yaml.representer import Representer
from ruamel.yaml.tokens import *

if False: ...

class YAML:
    def __init__(
        self: Any,
        *,
        typ: Optional[Union[list[str], str]] = ...,
        pure: Any = ...,
        output: Any = ...,
        plug_ins: Any = ...,
    ) -> None:
        """
        typ: 'rt'/None -> RoundTripLoader/RoundTripDumper,  (default)
             'safe'    -> SafeLoader/SafeDumper,
             'unsafe'  -> normal/unsafe Loader/Dumper (pending deprecation)
             'full'    -> full Dumper only, including python built-ins that are
                          potentially unsafe to load
             'base'    -> baseloader
        pure: if True only use Python modules
        input/output: needed to work as context manager
        plug_ins: a list of plug-in files
        """
        ...

    @property
    def reader(self) -> Any: ...
    @property
    def scanner(self) -> Any: ...
    @property
    def parser(self) -> Any: ...
    @property
    def composer(self) -> Any: ...
    @property
    def constructor(self) -> Any: ...
    @property
    def resolver(self) -> Any: ...
    @property
    def emitter(self) -> Any: ...
    @property
    def serializer(self) -> Any: ...
    @property
    def representer(self) -> Any: ...
    def scan(self, stream: StreamTextType) -> Any:
        """
        Scan a YAML stream and produce scanning tokens.
        """
        ...

    def parse(self, stream: StreamTextType) -> Any:
        """
        Parse a YAML stream and produce parsing events.
        """
        ...

    def compose(self, stream: Union[Path, StreamTextType]) -> Any:
        """
        Parse the first YAML document in a stream
        and produce the corresponding representation tree.
        """
        ...

    def compose_all(self, stream: Union[Path, StreamTextType]) -> Any:
        """
        Parse all YAML documents in a stream
        and produce corresponding representation trees.
        """
        ...

    def load(self, stream: Union[Path, StreamTextType]) -> Any:
        """
        at this point you either have the non-pure Parser (which has its own reader and
        scanner) or you have the pure Parser.
        If the pure Parser is set, then set the Reader and Scanner, if not already set.
        If either the Scanner or Reader are set, you cannot use the non-pure Parser,
            so reset it to the pure parser and set the Reader resp. Scanner if necessary
        """
        ...

    def load_all(self, stream: Union[Path, StreamTextType]) -> Any: ...
    def get_constructor_parser(self, stream: StreamTextType) -> Any:
        """
        the old cyaml needs special setup, and therefore the stream
        """
        ...

    def emit(self, events: Any, stream: Any) -> None:
        """
        Emit YAML parsing events into a stream.
        If stream is None, return the produced string instead.
        """
        ...

    def serialize(self, node: Any, stream: Optional[StreamType]) -> Any:
        """
        Serialize a representation tree into a YAML stream.
        If stream is None, return the produced string instead.
        """
        ...

    def serialize_all(self, nodes: Any, stream: Optional[StreamType]) -> Any:
        """
        Serialize a sequence of representation trees into a YAML stream.
        If stream is None, return the produced string instead.
        """
        ...

    def dump(
        self: Any,
        data: Union[Path, StreamType],
        stream: Any = ...,
        *,
        transform: Any = ...,
    ) -> Any: ...
    def dump_all(
        self, documents: Any, stream: Union[Path, StreamType], *, transform: Any = ...
    ) -> Any: ...
    def Xdump_all(self, documents: Any, stream: Any, *, transform: Any = ...) -> Any:
        """
        Serialize a sequence of Python objects into a YAML stream.
        """
        ...

    def get_serializer_representer_emitter(self, stream: StreamType, tlca: Any) -> Any:
        class XDumper(CEmitter, self.Representer, rslvr): ...

    def map(self, **kw: Any) -> Any: ...
    def seq(self, *args: Any) -> Any: ...
    def official_plug_ins(self) -> Any:
        """Search for list of subdirs that are plug-ins, if __file__ is not available, e.g.
        single file installers that are not properly emulating a file-system (issue 324)
        no plug-ins will be found. If any are packaged, you know which file that are
        and you can explicitly provide it during instantiation:
            yaml = ruamel.yaml.YAML(plug_ins=['ruamel/yaml/jinja2/__plug_in__'])
        """
        ...

    def register_class(self, cls: Any) -> Any:
        """
        register a class for dumping/loading
        - if it has attribute yaml_tag use that to register, else use class name
        - if it has methods to_yaml/from_yaml use those to dump/load else dump attributes
          as mapping
        """
        ...

    def __enter__(self) -> Any: ...
    def __exit__(
        self,
        typ: Optional[type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None: ...
    @property
    def version(self) -> Optional[tuple[int, int]]: ...
    @version.setter
    def version(self, val: VersionType) -> None: ...
    @property
    def tags(self) -> Any: ...
    @tags.setter
    def tags(self, val: Any) -> None: ...
    @property
    def indent(self) -> Any: ...
    @indent.setter
    def indent(self, val: Any) -> None: ...
    @property
    def block_seq_indent(self) -> Any: ...
    @block_seq_indent.setter
    def block_seq_indent(self, val: Any) -> None: ...
    def compact(self, seq_seq: Any = ..., seq_map: Any = ...) -> None: ...

class YAMLContextManager:
    def __init__(self, yaml: Any, transform: Any = ...) -> None: ...
    def teardown_output(self) -> None: ...
    def init_output(self, first_data: Any) -> None: ...
    def dump(self, data: Any) -> None: ...

def yaml_object(yml: Any) -> Any:
    """Decorator for classes that needs to dump/load objects
    The tag for such objects is taken from the class attribute yaml_tag (or the
    class name in lowercase in case unavailable)
    If methods to_yaml and/or from_yaml are available, these are called for dumping resp.
    loading, default routines (dumping a mapping of the attributes) used otherwise.
    """
    ...

def warn_deprecation(fun: Any, method: Any, arg: str = ...) -> None: ...
def error_deprecation(
    fun: Any, method: Any, arg: str = ..., comment: str = ...
) -> None: ...

_error_dep_arg = ...
_error_dep_comment = ...

def scan(stream: StreamTextType, Loader: Any = ...) -> Any:
    """
    Scan a YAML stream and produce scanning tokens.
    """
    ...

def parse(stream: StreamTextType, Loader: Any = ...) -> Any:
    """
    Parse a YAML stream and produce parsing events.
    """
    ...

def compose(stream: StreamTextType, Loader: Any = ...) -> Any:
    """
    Parse the first YAML document in a stream
    and produce the corresponding representation tree.
    """
    ...

def compose_all(stream: StreamTextType, Loader: Any = ...) -> Any:
    """
    Parse all YAML documents in a stream
    and produce corresponding representation trees.
    """
    ...

def load(
    stream: Any, Loader: Any = ..., version: Any = ..., preserve_quotes: Any = ...
) -> Any:
    """
    Parse the first YAML document in a stream
    and produce the corresponding Python object.
    """
    ...

def load_all(
    stream: Any, Loader: Any = ..., version: Any = ..., preserve_quotes: Any = ...
) -> Any:
    """
    Parse all YAML documents in a stream
    and produce corresponding Python objects.
    """
    ...

def safe_load(stream: StreamTextType, version: Optional[VersionType] = ...) -> Any:
    """
    Parse the first YAML document in a stream
    and produce the corresponding Python object.
    Resolve only basic YAML tags.
    """
    ...

def safe_load_all(stream: StreamTextType, version: Optional[VersionType] = ...) -> Any:
    """
    Parse all YAML documents in a stream
    and produce corresponding Python objects.
    Resolve only basic YAML tags.
    """
    ...

def round_trip_load(
    stream: StreamTextType,
    version: Optional[VersionType] = ...,
    preserve_quotes: Optional[bool] = ...,
) -> Any:
    """
    Parse the first YAML document in a stream
    and produce the corresponding Python object.
    Resolve only basic YAML tags.
    """
    ...

def round_trip_load_all(
    stream: StreamTextType,
    version: Optional[VersionType] = ...,
    preserve_quotes: Optional[bool] = ...,
) -> Any:
    """
    Parse all YAML documents in a stream
    and produce corresponding Python objects.
    Resolve only basic YAML tags.
    """
    ...

def emit(
    events: Any,
    stream: Optional[StreamType] = ...,
    Dumper: Any = ...,
    canonical: Optional[bool] = ...,
    indent: Union[int, None] = ...,
    width: Optional[int] = ...,
    allow_unicode: Optional[bool] = ...,
    line_break: Any = ...,
) -> Any:
    """
    Emit YAML parsing events into a stream.
    If stream is None, return the produced string instead.
    """
    ...

enc = ...

def serialize_all(
    nodes: Any,
    stream: Optional[StreamType] = ...,
    Dumper: Any = ...,
    canonical: Any = ...,
    indent: Optional[int] = ...,
    width: Optional[int] = ...,
    allow_unicode: Optional[bool] = ...,
    line_break: Any = ...,
    encoding: Any = ...,
    explicit_start: Optional[bool] = ...,
    explicit_end: Optional[bool] = ...,
    version: Optional[VersionType] = ...,
    tags: Any = ...,
) -> Any:
    """
    Serialize a sequence of representation trees into a YAML stream.
    If stream is None, return the produced string instead.
    """
    ...

def serialize(
    node: Any, stream: Optional[StreamType] = ..., Dumper: Any = ..., **kwds: Any
) -> Any:
    """
    Serialize a representation tree into a YAML stream.
    If stream is None, return the produced string instead.
    """
    ...

def dump_all(
    documents: Any,
    stream: Optional[StreamType] = ...,
    Dumper: Any = ...,
    default_style: Any = ...,
    default_flow_style: Any = ...,
    canonical: Optional[bool] = ...,
    indent: Optional[int] = ...,
    width: Optional[int] = ...,
    allow_unicode: Optional[bool] = ...,
    line_break: Any = ...,
    encoding: Any = ...,
    explicit_start: Optional[bool] = ...,
    explicit_end: Optional[bool] = ...,
    version: Any = ...,
    tags: Any = ...,
    block_seq_indent: Any = ...,
    top_level_colon_align: Any = ...,
    prefix_colon: Any = ...,
) -> Any:
    """
    Serialize a sequence of Python objects into a YAML stream.
    If stream is None, return the produced string instead.
    """
    ...

def dump(
    data: Any,
    stream: Optional[StreamType] = ...,
    Dumper: Any = ...,
    default_style: Any = ...,
    default_flow_style: Any = ...,
    canonical: Optional[bool] = ...,
    indent: Optional[int] = ...,
    width: Optional[int] = ...,
    allow_unicode: Optional[bool] = ...,
    line_break: Any = ...,
    encoding: Any = ...,
    explicit_start: Optional[bool] = ...,
    explicit_end: Optional[bool] = ...,
    version: Optional[VersionType] = ...,
    tags: Any = ...,
    block_seq_indent: Any = ...,
) -> Any:
    """
    Serialize a Python object into a YAML stream.
    If stream is None, return the produced string instead.

    default_style ∈ None, '', '"', "'", '|', '>'

    """
    ...

def safe_dump(data: Any, stream: Optional[StreamType] = ..., **kwds: Any) -> Any:
    """
    Serialize a Python object into a YAML stream.
    Produce only basic YAML tags.
    If stream is None, return the produced string instead.
    """
    ...

def round_trip_dump(
    data: Any,
    stream: Optional[StreamType] = ...,
    Dumper: Any = ...,
    default_style: Any = ...,
    default_flow_style: Any = ...,
    canonical: Optional[bool] = ...,
    indent: Optional[int] = ...,
    width: Optional[int] = ...,
    allow_unicode: Optional[bool] = ...,
    line_break: Any = ...,
    encoding: Any = ...,
    explicit_start: Optional[bool] = ...,
    explicit_end: Optional[bool] = ...,
    version: Optional[VersionType] = ...,
    tags: Any = ...,
    block_seq_indent: Any = ...,
    top_level_colon_align: Any = ...,
    prefix_colon: Any = ...,
) -> Any: ...
def add_implicit_resolver(
    tag: Any,
    regexp: Any,
    first: Any = ...,
    Loader: Any = ...,
    Dumper: Any = ...,
    resolver: Any = ...,
) -> None:
    """
    Add an implicit scalar detector.
    If an implicit scalar value matches the given regexp,
    the corresponding tag is assigned to the scalar.
    first is a sequence of possible initial characters or None.
    """
    ...

def add_path_resolver(
    tag: Any,
    path: Any,
    kind: Any = ...,
    Loader: Any = ...,
    Dumper: Any = ...,
    resolver: Any = ...,
) -> None:
    """
    Add a path based resolver for the given tag.
    A path is a list of keys that forms a path
    to a node in the representation tree.
    Keys can be string values, integers, or None.
    """
    ...

def add_constructor(
    tag: Any, object_constructor: Any, Loader: Any = ..., constructor: Any = ...
) -> None:
    """
    Add an object constructor for the given tag.
    object_onstructor is a function that accepts a Loader instance
    and a node object and produces the corresponding Python object.
    """
    ...

def add_multi_constructor(
    tag_prefix: Any, multi_constructor: Any, Loader: Any = ..., constructor: Any = ...
) -> None:
    """
    Add a multi-constructor for the given tag prefix.
    Multi-constructor is called for a node if its tag starts with tag_prefix.
    Multi-constructor accepts a Loader instance, a tag suffix,
    and a node object and produces the corresponding Python object.
    """
    ...

def add_representer(
    data_type: Any, object_representer: Any, Dumper: Any = ..., representer: Any = ...
) -> None:
    """
    Add a representer for the given type.
    object_representer is a function accepting a Dumper instance
    and an instance of the given data type
    and producing the corresponding representation node.
    """
    ...

def add_multi_representer(
    data_type: Any, multi_representer: Any, Dumper: Any = ..., representer: Any = ...
) -> None:
    """
    Add a representer for the given type.
    multi_representer is a function accepting a Dumper instance
    and an instance of the given data type or subtype
    and producing the corresponding representation node.
    """
    ...

class YAMLObjectMetaclass(type):
    """
    The metaclass for YAMLObject.
    """

    def __init__(cls, name: Any, bases: Any, kwds: Any) -> None: ...

class YAMLObject(with_metaclass(YAMLObjectMetaclass)):
    """
    An object that can dump itself to a YAML stream
    and load itself from a YAML stream.
    """

    __slots__ = ...
    yaml_constructor = Constructor
    yaml_representer = Representer
    yaml_tag: Any = ...
    yaml_flow_style: Any = ...
    @classmethod
    def from_yaml(cls, constructor: Any, node: Any) -> Any:
        """
        Convert a representation node to a Python object.
        """
        ...

    @classmethod
    def to_yaml(cls, representer: Any, data: Any) -> Any:
        """
        Convert a Python object to a representation node.
        """
        ...



import collections
from collections import OrderedDict
from pathlib import Path, PosixPath, WindowsPath
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ruamel.yaml import YAML, Dumper
else:
    try:
        from ruamel.yaml import YAML, Dumper
    except ImportError:  # pragma: no cover
        from ruamel_yaml import YAML, Dumper


def get_yaml_io():
    """Return yaml loader/dumper configured for snakebids."""
    yaml = YAML()

    # Represent any PathLikes as str.
    def path2str(dumper: Dumper, data: Path):
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str",
            str(data),
        )

    def to_dict(dumper: Dumper, data: OrderedDict[Any, Any]):
        return dumper.represent_dict(dict(data))

    yaml.representer.add_representer(PosixPath, path2str)
    yaml.representer.add_representer(WindowsPath, path2str)
    yaml.representer.add_representer(collections.OrderedDict, to_dict)
    yaml.representer.add_representer(OrderedDict, to_dict)
    return yaml



from __future__ import annotations

import platform
import re
import subprocess as sp
import sys
import tempfile
from pathlib import Path
from typing import Any, Final, Literal, TypedDict

import copier
import more_itertools as itx
import pathvalidate
import pytest
import requests_mock
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from pytest_mock.plugin import MockerFixture
from typing_extensions import NotRequired, Unpack

from snakebids.jinja2_ext.format_dep_spec import format_poetry
from snakebids.jinja2_ext.snakebids_version import impm

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

import snakebids
from tests.helpers import allow_function_scoped, needs_docker

BuildBackend = Literal["poetry", "hatch", "flit", "setuptools"]
BUILD_BACKENDS: Final[list[BuildBackend]] = ["poetry", "hatch", "flit", "setuptools"]


class DataFields(TypedDict):
    full_name: str
    email: str
    app_full_name: str
    github: str
    app_description: str
    build_system: BuildBackend
    app_version: str
    create_doc_template: bool
    license: str
    snakebids_version: NotRequired[str]


def get_empty_data(app_name: str, build: BuildBackend) -> DataFields:
    data: DataFields = {
        "full_name": "",
        "email": "",
        "app_full_name": app_name,
        "github": "",
        "app_description": "",
        "build_system": build,
        "app_version": "0.1.0",
        "create_doc_template": False,
        "license": "",
    }
    # poetry we need an author
    if build == "poetry":
        data["full_name"] = "John Doe"
        data["email"] = "example@email.com"
    return data


# emails complements of https://gist.github.com/cjaoude/fd9910626629b53c4d25
def invalid_emails():
    return st.sampled_from(
        [
            "plainaddress",
            "#@%^%#$@#$@#.com",
            "@example.com",
            "Joe Smith <email@example.com>",
            "email.example.com",
            "email@example@example.com",
            ".email@example.com",
            "email.@example.com",
            "email..email@example.com",
            "あいうえお@example.com",
            "email@example.com (Joe Smith)",
            "email@example",
            "email@-example.com",
            "email@example.web-",
            "email@[111.123.123.4444]",
            "email@example..com",
            "Abc..123@example.com",
            r"”(),:;<>[\]@example.com",
            "just”not”right@example.com",
            r'this\ is"really"not\allowed@example.com',
        ]
    )


@given(email=invalid_emails())
@allow_function_scoped
def test_invalid_email_raises_error(email: str, tmp_path: Path):
    data = get_empty_data("testapp", "setuptools")
    data["email"] = email
    with pytest.raises(ValueError, match="Must be a valid email"):
        copier.run_copy(
            str(Path(itx.first(snakebids.__path__), "project_template")),
            tmp_path / data["app_full_name"],
            data=data,
            unsafe=True,
        )


@pytest.mark.parametrize("spec_type", ["version", "url", "path", "git"])
@given(spec=st.text().map(lambda s: s.strip()).filter(lambda s: not s.startswith("@")))
def test_format_poetry(spec_type: Literal["version", "url", "path", "git"], spec: str):
    prefix = {"url": "@ ", "path": "@ file://", "git": "@ git+", "version": ""}[
        spec_type
    ]
    if spec_type == "url":
        assume("file://" not in spec and "git+" not in spec)
    parsed = tomllib.loads("spec = " + format_poetry(prefix + spec))["spec"]
    if spec_type == "version":
        assert parsed == spec
    else:
        assert parsed[spec_type] == spec


@pytest.mark.parametrize("build", BUILD_BACKENDS)
@pytest.mark.parametrize(
    ("server", "server_status", "metadata", "expected"),
    [
        # Server returns valid version
        ("1.1.1", 200, "2.2.2", "1.1.1"),
        # Server raises 400 error
        ("1.1.1", 400, "2.2.2", "2.2.2"),
        # server returns invalid and importlib.metadata gives valid
        ("1.1.1.dev1", 200, "2.2.2", "2.2.2"),
        # server invalid and importlib.metadata gives 0.0.0
        ("1.1.1.dev1", 200, "0.0.0", "0.12.0"),
        # server and importlib.metadata invalid
        ("1.1.1.dev1", 200, "2.2.2.dev1", "0.12.0"),
        # server gives non-string type
        (True, 200, "2.2.2", "2.2.2"),
    ],
)
def test_gets_correct_snakebids_version(
    build: BuildBackend,
    server: Any,
    server_status: str,
    metadata: str,
    expected: str,
    mocker: MockerFixture,
    tmp_path: Path,
):
    tmpdir = Path(tempfile.mkdtemp(dir=tmp_path))
    with requests_mock.Mocker() as m:
        m.get(
            "https://pypi.org/pypi/snakebids/json",
            json={"info": {"version": server}},
            status_code=server_status,
        )

        mocker.patch.object(impm, "version", return_value=metadata)
        data = get_empty_data("testapp", build)
        copier.run_copy(
            str(Path(itx.first(snakebids.__path__), "project_template")),
            tmpdir / data["app_full_name"],
            data=data,
            unsafe=True,
        )
    with open(tmpdir / data["app_full_name"] / "pyproject.toml", "rb") as f:
        pyproject = tomllib.load(f)
    if build == "poetry":
        assert (
            pyproject["tool"]["poetry"]["dependencies"]["snakebids"] == f">={expected}"
        )
    else:
        assert f"snakebids >={expected}" in pyproject["project"]["dependencies"]


@given(
    name=st.text()
    .filter(lambda s: not re.match(r"^[a-zA-Z_][a-zA-Z_0-9]*$", s))
    .filter(lambda s: pathvalidate.is_valid_filename(s, pathvalidate.Platform.LINUX))
)
@allow_function_scoped
def test_invalid_app_name_raises_error(name: str, tmp_path: Path):
    data = get_empty_data(name, "setuptools")
    with pytest.raises(ValueError, match="Name must be a valid"):
        copier.run_copy(
            str(Path(itx.first(snakebids.__path__), "project_template")),
            tmp_path / data["app_full_name"],
            data=data,
            unsafe=True,
        )


@pytest.mark.parametrize(
    ("build", "build_backend"),
    [
        ("poetry", "poetry.core.masonry.api"),
        ("hatch", "hatchling.build"),
        ("flit", "flit_core.buildapi"),
        ("setuptools", "setuptools.build_meta"),
    ],
)
def test_correct_build_system_used(
    tmp_path: Path, build: BuildBackend, build_backend: str
):
    tmpdir = Path(tempfile.mkdtemp(dir=tmp_path))
    data = get_empty_data("testapp", build)
    copier.run_copy(
        str(Path(itx.first(snakebids.__path__), "project_template")),
        tmpdir / data["app_full_name"],
        data=data,
        unsafe=True,
    )
    with open(tmpdir / data["app_full_name"] / "pyproject.toml", "rb") as f:
        pyproject = tomllib.load(f)
    assert pyproject["build-system"]["build-backend"] == build_backend


@given(
    full_name=st.text(),
    email=st.emails() | st.just(""),
    app_full_name=st.from_regex(r"[a-zA-Z_][a-zA-Z_0-9]*", fullmatch=True),
    github=st.text(),
    app_description=st.text(),
    app_version=st.text(min_size=1),
    create_doc_template=st.just(False),
    license=st.text(),
    snakebids_version=st.text().filter(lambda s: not s.strip().startswith("@")),
)
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture], deadline=5000)
@pytest.mark.parametrize("build", BUILD_BACKENDS)
def test_pyproject_correctly_formatted(
    tmp_path: Path, build: BuildBackend, **kwargs: Unpack[DataFields]
):
    tmpdir = Path(tempfile.mkdtemp(dir=tmp_path))
    kwargs["build_system"] = build
    copier.run_copy(
        str(Path(itx.first(snakebids.__path__), "project_template")),
        tmpdir / kwargs["app_full_name"],
        data=kwargs,
        unsafe=True,
    )
    with open(tmpdir / kwargs["app_full_name"] / "pyproject.toml", "rb") as f:
        pyproject = tomllib.load(f)
    if build == "poetry":
        assert kwargs["app_full_name"] == pyproject["tool"]["poetry"]["name"]
        assert kwargs["app_version"] == pyproject["tool"]["poetry"]["version"]
        assert kwargs["app_description"] == pyproject["tool"]["poetry"]["description"]
        if kwargs["license"]:
            assert kwargs["license"] == pyproject["tool"]["poetry"]["license"]
        else:
            assert "license" not in pyproject["tool"]["poetry"]
        if kwargs["full_name"]:
            assert len(pyproject["tool"]["poetry"]["authors"]) == 1
            email_tag = f" <{kwargs['email']}>" if kwargs["email"] else ""
            assert (
                f"{kwargs['full_name']}{email_tag}"
                == pyproject["tool"]["poetry"]["authors"][0]
            )
        else:
            assert "authors" not in pyproject["tool"]["poetry"]
        assert (
            pyproject["tool"]["poetry"]["dependencies"]["snakebids"]
            == kwargs["snakebids_version"]  # type: ignore
        )
        return

    assert kwargs["app_full_name"] == pyproject["project"]["name"]
    assert kwargs["app_version"] == pyproject["project"]["version"]
    assert kwargs["app_description"] == pyproject["project"]["description"]
    if kwargs["license"]:
        assert kwargs["license"] == pyproject["project"]["license"]
    else:
        assert "license" not in pyproject["project"]
    if kwargs["full_name"] or kwargs["email"]:
        assert len(pyproject["project"]["authors"]) == 1
        author_obj: dict[str, str] = {}
        if kwargs["full_name"]:
            author_obj["name"] = kwargs["full_name"]
        if kwargs["email"]:
            author_obj["email"] = kwargs["email"]
        assert author_obj == pyproject["project"]["authors"][0]
    else:
        assert "authors" not in pyproject["project"]
    pyproject["project"]["dependencies"].index(
        f"snakebids {kwargs['snakebids_version']}"  # type: ignore
    )


@needs_docker(f"snakebids/test-template:{platform.python_version()}")
@pytest.mark.parametrize(
    ("build", "venv"),
    [
        ("setuptools", "setuptools"),
        ("poetry", "poetry"),
        ("hatch", "hatch"),
        ("flit", "setuptools"),
    ],
)
def test_template_dry_runs_successfully(
    tmp_path: Path, request: pytest.FixtureRequest, build: BuildBackend, venv: str
):
    app_name = "snakebids_app"
    data = get_empty_data(app_name, build)
    data["snakebids_version"] = "@ file:///src"

    copier.run_copy(
        str(Path(itx.first(snakebids.__path__), "project_template")),
        tmp_path / app_name,
        data=data,
        unsafe=True,
    )
    cmd = sp.run(
        [
            "docker",
            "run",
            "-v",
            f"{tmp_path / app_name}:/app",
            "-v",
            f"{request.config.rootpath}:/snakebids",
            "--rm",
            f"snakebids/test-template:{platform.python_version()}",
            venv,
            app_name,
        ],
        capture_output=True,
        check=False,
    )
    try:
        cmd.check_returncode()
    except Exception:
        print(cmd.stdout.decode())
        print(cmd.stderr.decode(), file=sys.stderr)
        raise
    assert "All set" in cmd.stdout.decode()


@needs_docker(f"snakebids/test-template:{platform.python_version()}")
def test_template_docs_build(tmp_path: Path, request: pytest.FixtureRequest):
    app_name = "snakebids_app"
    data = get_empty_data(app_name, "setuptools")
    data["create_doc_template"] = True
    data["snakebids_version"] = "@ file:///src"

    copier.run_copy(
        str(Path(itx.first(snakebids.__path__), "project_template")),
        tmp_path / app_name,
        data=data,
        unsafe=True,
    )
    cmd = sp.run(
        [
            "docker",
            "run",
            "-v",
            f"{tmp_path / app_name}:/app",
            "-v",
            f"{request.config.rootpath}:/snakebids",
            "--rm",
            f"snakebids/test-template:{platform.python_version()}",
            "docs",
            app_name,
        ],
        capture_output=True,
        check=False,
    )
    try:
        cmd.check_returncode()
    except Exception:
        print(cmd.stdout.decode())
        print(cmd.stderr.decode(), file=sys.stderr)
        raise



"""
This type stub file was generated by pyright.
"""

"""File-indexing functionality. """

class BIDSLayoutIndexer:
    """Indexer class for BIDSLayout.

    Parameters
    ----------
    validate : bool, optional
        If True, all files are checked for BIDS compliance when first indexed,
        and non-compliant files are ignored. This provides a convenient way to
        restrict file indexing to only those files defined in the "core" BIDS
        spec, as setting ``validate=True`` will lead noncompliant files
        like ``sub-01/nonbidsfile.txt`` to be ignored.
    ignore : str or SRE_Pattern or list
        Path(s) to exclude from indexing. Each path is either a string or a
        SRE_Pattern object (i.e., compiled regular expression). If a string is
        passed, it must be either an absolute path, or be relative to the BIDS
        project root. If an SRE_Pattern is passed, the contained regular
        expression will be matched against the full (absolute) path of all
        files and directories. By default, indexing ignores all files in
        'code/', 'stimuli/', 'sourcedata/', 'models/', and any hidden
        files/dirs beginning with '.' at root level.
    force_index : str or SRE_Pattern or list
        Path(s) to forcibly index in the BIDSLayout, even if they would
        otherwise fail validation. See the documentation for the ignore
        argument for input format details. Note that paths in force_index takes
        precedence over those in ignore (i.e., if a file matches both ignore
        and force_index, it *will* be indexed).
        Note: NEVER include 'derivatives' here; use the derivatives argument
        (or :obj:`bids.layout.BIDSLayout.add_derivatives`) for that.
    index_metadata : bool
        If True, all metadata files are indexed. If False, metadata will not be
        available (but indexing will be faster).
    config_filename : str
        Optional name of filename within directories
        that contains configuration information.
    **filters
        keyword arguments passed to the .get() method of a
        :obj:`bids.layout.BIDSLayout` object. These keyword arguments define
        what files get selected for metadata indexing.
    """

    def __init__(
        self,
        validate=...,
        ignore=...,
        force_index=...,
        index_metadata=...,
        config_filename=...,
        **filters
    ) -> None: ...
    def __call__(self, layout): ...
    @property
    def session(self): ...



"""
This type stub file was generated by pyright.
"""

import datetime
from ruamel.yaml.error import MarkedYAMLError, MarkedYAMLFutureWarning
from ruamel.yaml.nodes import *
from ruamel.yaml.comments import *
from ruamel.yaml.comments import (
    CommentedMap,
    CommentedOrderedMap,
    CommentedSeq,
    CommentedSet,
    TaggedScalar,
)
from ruamel.yaml.scalarbool import ScalarBoolean
from ruamel.yaml.timestamp import TimeStamp
from ruamel.yaml.util import timestamp_regexp

if False: ...
__all__ = [
    "BaseConstructor",
    "SafeConstructor",
    "Constructor",
    "ConstructorError",
    "RoundTripConstructor",
]

class ConstructorError(MarkedYAMLError): ...
class DuplicateKeyFutureWarning(MarkedYAMLFutureWarning): ...
class DuplicateKeyError(MarkedYAMLError): ...

class BaseConstructor:
    yaml_constructors: Dict[Any, Any] = ...
    yaml_multi_constructors: Dict[Any, Any] = ...
    def __init__(
        self, preserve_quotes: Optional[bool] = ..., loader: Any = ...
    ) -> None: ...
    @property
    def composer(self) -> Any: ...
    @property
    def resolver(self) -> Any: ...
    @property
    def scanner(self) -> Any: ...
    def check_data(self) -> Any: ...
    def get_data(self) -> Any: ...
    def get_single_data(self) -> Any: ...
    def construct_document(self, node: Any) -> Any: ...
    def construct_object(self, node: Any, deep: bool = ...) -> Any:
        """deep is True when creating an object/mapping recursively,
        in that case want the underlying elements available during construction
        """
        ...

    def construct_non_recursive_object(
        self, node: Any, tag: Optional[str] = ...
    ) -> Any: ...
    def construct_scalar(self, node: Any) -> Any: ...
    def construct_sequence(self, node: Any, deep: bool = ...) -> Any:
        """deep is True when creating an object/mapping recursively,
        in that case want the underlying elements available during construction
        """
        ...

    def construct_mapping(self, node: Any, deep: bool = ...) -> Any:
        """deep is True when creating an object/mapping recursively,
        in that case want the underlying elements available during construction
        """
        ...

    def check_mapping_key(
        self, node: Any, key_node: Any, mapping: Any, key: Any, value: Any
    ) -> bool:
        """return True if key is unique"""
        ...

    def check_set_key(
        self: Any, node: Any, key_node: Any, setting: Any, key: Any
    ) -> None: ...
    def construct_pairs(self, node: Any, deep: bool = ...) -> Any: ...
    @classmethod
    def add_constructor(cls, tag: Any, constructor: Any) -> Any: ...
    @classmethod
    def add_multi_constructor(cls, tag_prefix: Any, multi_constructor: Any) -> None: ...
    @classmethod
    def add_default_constructor(
        cls, tag: str, method: Any = ..., tag_base: str = ...
    ) -> None: ...

class SafeConstructor(BaseConstructor):
    def construct_scalar(self, node: Any) -> Any: ...
    def flatten_mapping(self, node: Any) -> Any:
        """
        This implements the merge key feature http://yaml.org/type/merge.html
        by inserting keys from the merge dict/list of dicts if not yet
        available in this node
        """
        ...

    def construct_mapping(self, node: Any, deep: bool = ...) -> Any:
        """deep is True when creating an object/mapping recursively,
        in that case want the underlying elements available during construction
        """
        ...

    def construct_yaml_null(self, node: Any) -> Any: ...

    bool_values = ...
    def construct_yaml_bool(self, node: Any) -> bool: ...
    def construct_yaml_int(self, node: Any) -> int: ...

    inf_value = ...
    nan_value = ...
    def construct_yaml_float(self, node: Any) -> float: ...
    def construct_yaml_binary(self, node: Any) -> Any: ...

    timestamp_regexp = ...
    def construct_yaml_timestamp(self, node: Any, values: Any = ...) -> Any: ...
    def construct_yaml_omap(self, node: Any) -> Any: ...
    def construct_yaml_pairs(self, node: Any) -> Any: ...
    def construct_yaml_set(self, node: Any) -> Any: ...
    def construct_yaml_str(self, node: Any) -> Any: ...
    def construct_yaml_seq(self, node: Any) -> Any: ...
    def construct_yaml_map(self, node: Any) -> Any: ...
    def construct_yaml_object(self, node: Any, cls: Any) -> Any: ...
    def construct_undefined(self, node: Any) -> None: ...

class Constructor(SafeConstructor):
    def construct_python_str(self, node: Any) -> Any: ...
    def construct_python_unicode(self, node: Any) -> Any: ...
    def construct_python_bytes(self, node: Any) -> Any: ...
    def construct_python_long(self, node: Any) -> int: ...
    def construct_python_complex(self, node: Any) -> Any: ...
    def construct_python_tuple(self, node: Any) -> Any: ...
    def find_python_module(self, name: Any, mark: Any) -> Any: ...
    def find_python_name(self, name: Any, mark: Any) -> Any: ...
    def construct_python_name(self, suffix: Any, node: Any) -> Any: ...
    def construct_python_module(self, suffix: Any, node: Any) -> Any: ...
    def make_python_instance(
        self,
        suffix: Any,
        node: Any,
        args: Any = ...,
        kwds: Any = ...,
        newobj: bool = ...,
    ) -> Any: ...
    def set_python_instance_state(self, instance: Any, state: Any) -> None: ...
    def construct_python_object(self, suffix: Any, node: Any) -> Any: ...
    def construct_python_object_apply(
        self, suffix: Any, node: Any, newobj: bool = ...
    ) -> Any: ...
    def construct_python_object_new(self, suffix: Any, node: Any) -> Any: ...
    @classmethod
    def add_default_constructor(
        cls, tag: str, method: Any = ..., tag_base: str = ...
    ) -> None: ...

class RoundTripConstructor(SafeConstructor):
    """need to store the comments on the node itself,
    as well as on the items
    """
    def comment(self, idx: Any) -> Any: ...
    def comments(self, list_of_comments: Any, idx: Optional[Any] = ...) -> Any: ...
    def construct_scalar(self, node: Any) -> Any: ...
    def construct_yaml_int(self, node: Any) -> Any: ...
    def construct_yaml_float(self, node: Any) -> Any: ...
    def construct_yaml_str(self, node: Any) -> Any: ...
    def construct_rt_sequence(
        self, node: Any, seqtyp: Any, deep: bool = ...
    ) -> Any: ...
    def flatten_mapping(self, node: Any) -> Any:
        """
        This implements the merge key feature http://yaml.org/type/merge.html
        by inserting keys from the merge dict/list of dicts if not yet
        available in this node
        """
        ...

    def construct_mapping(self, node: Any, maptyp: Any, deep: bool = ...) -> Any: ...
    def construct_setting(self, node: Any, typ: Any, deep: bool = ...) -> Any: ...
    def construct_yaml_seq(self, node: Any) -> Iterator[CommentedSeq]: ...
    def construct_yaml_map(self, node: Any) -> Iterator[CommentedMap]: ...
    def set_collection_style(self, data: Any, node: Any) -> None: ...
    def construct_yaml_object(self, node: Any, cls: Any) -> Any: ...
    def construct_yaml_omap(self, node: Any) -> Iterator[CommentedOrderedMap]: ...
    def construct_yaml_set(self, node: Any) -> Iterator[CommentedSet]: ...
    def construct_unknown(
        self, node: Any
    ) -> Iterator[Union[CommentedMap, TaggedScalar, CommentedSeq]]: ...
    def construct_yaml_timestamp(
        self, node: Any, values: Any = ...
    ) -> Union[datetime.date, datetime.datetime, TimeStamp]: ...
    def construct_yaml_sbool(self, node: Any) -> Union[bool, ScalarBoolean]: ...



from __future__ import annotations

import copy
import itertools as it
from collections.abc import Container, Hashable, Iterable, Sequence
from os import PathLike
from pathlib import Path
from string import ascii_letters, digits
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

import hypothesis.strategies as st
from bids.layout import Config as BidsConfig
from hypothesis import assume

from snakebids.core.datasets import (
    BidsComponent,
    BidsComponentRow,
    BidsDataset,
    BidsPartialComponent,
)
from snakebids.types import Expandable, InputConfig, InputsConfig, ZipList
from snakebids.utils.containers import ContainerBag, MultiSelectDict
from snakebids.utils.utils import BidsEntity
from tests import helpers

_Ex_co = TypeVar("_Ex_co", bound=str, covariant=True)
_T = TypeVar("_T")

alphanum = ascii_letters + digits
valid_entities: tuple[str, ...] = tuple(BidsConfig.load("bids").entities.keys())

path_characters = st.characters(blacklist_characters=["/", "\x00"], codec="UTF-8")
# StackOverflow answer by Gumbo
# https://stackoverflow.com/questions/1547899/which-characters-make-a-url-invalid#1547940
scheme_characters = st.sampled_from(
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~?#[]@!$&'()*+,;="
)


def nothing() -> Any:
    return st.nothing()  # type: ignore


def schemes() -> st.SearchStrategy[str]:
    # Generate the prefix part of a url.
    random_st = st.text(scheme_characters, min_size=1).map(lambda s: f"{s}://")
    fixed_st = st.sampled_from(["C://", "c:\\", "gs://", "s3://"])
    return random_st | fixed_st


def paths(
    *,
    min_segments: int = 0,
    max_segments: int | None = None,
    absolute: bool | None = None,
    resolve: bool = False,
) -> st.SearchStrategy[Path]:
    paths = st.lists(
        st.text(path_characters, min_size=1),
        min_size=min_segments,
        max_size=max_segments,
    ).map(lambda x: Path(*x))

    relative_paths = paths.filter(lambda p: not p.is_absolute())

    absolute_paths = paths.map(lambda p: Path("/", p))

    if absolute:
        result = absolute_paths
    elif absolute is False:
        result = relative_paths
    else:
        result = absolute_paths | relative_paths

    if resolve:
        return result.map(lambda p: p.resolve())

    return result


def schemed_paths(
    *,
    min_segments: int = 0,
    max_segments: int | None = None,
) -> st.SearchStrategy[str]:
    """schemed paths are those beginning with scheme://.

    These paths are incompatible with paths() because pathlib.Path turns // into /
    when parsing, and paths() returns Path objects by default. That's why we have a
    separate function for schemed paths.

    This function returns un-schemed paths as well, occasionally
    """
    scheme_st: st.SearchStrategy[str] = (st.none() | schemes()).map(
        lambda s: "" if s is None else s
    )
    paths_st: st.SearchStrategy[str] = paths(
        min_segments=min_segments,
        max_segments=max_segments,
        absolute=False,
    ).map(lambda p: str(p))
    return st.tuples(scheme_st, paths_st).map(lambda t: t[0] + t[1])


def bids_entity(
    *,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    path_safe: bool = False,
) -> st.SearchStrategy[BidsEntity]:
    blacklist = ContainerBag(
        {"fmap", "scans"},
        blacklist_entities if blacklist_entities is not None else set(),
        {"datatype", "suffix", "extension"} if path_safe else set(),
    )
    return st.sampled_from(
        [
            BidsEntity(key)
            for key in valid_entities
            if key not in blacklist
            and (not whitelist_entities or key in whitelist_entities)
        ],
    )


def bids_value(pattern: str = r"[^\n\r\0]*") -> st.SearchStrategy[str]:
    # Need the isdigit == isdecimal check to work around a pybids bug
    return (
        st.from_regex(pattern, fullmatch=True)
        .filter(len)
        .filter(lambda s: all(c.isdigit() == c.isdecimal() for c in s))
        # remove values leading with multiple dots ".." because pybids doesn't handle
        # them correctly when applied to the "extension" entity
        .filter(lambda s: s[:2] != "..")
    )


def _filter_invalid_entity_lists(entities: Sequence[BidsEntity | str]):
    """Entity lists may not consist of just a datatype or extension.

    If suffix is in the entity list, so must extension
    """
    return all(
        [
            (
                # If suffix is in the path, extension must be too
                "suffix" not in entities or "extension" in entities
            ),
            # Cannot have paths with just datatype, just extension, or just datatype and
            # extension
            set(map(str, entities))
            not in [
                {"datatype"},
                {"extension"},
                {"datatype", "extension"},
            ],  # type: ignore
        ]
    )


@st.composite
def bids_path(
    draw: st.DrawFn,
    *,
    root: PathLike[str] | str | None = None,
    entities: Iterable[BidsEntity | str] | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    extra_entities: bool = True,
) -> Path:
    entities = (
        draw(
            bids_entity_lists(
                whitelist_entities=whitelist_entities,
                blacklist_entities=blacklist_entities,
            )
        )
        if entities is None
        else list(entities)
    )

    extras = (
        {
            k: v[0].replace("{", "{{").replace("}", "}}")
            for k, v in draw(
                zip_lists(
                    max_values=1,
                    max_entities=2,
                    blacklist_entities=ContainerBag(
                        blacklist_entities if blacklist_entities is not None else [],
                        [BidsEntity.normalize(e) for e in entities],
                    ),
                    restrict_patterns=True,
                )
            ).items()
        }
        if extra_entities
        else {}
    )

    return Path(root or ".") / helpers.get_bids_path(entities, **extras)


def bids_entity_lists(
    *,
    min_size: int = 1,
    max_size: int = 5,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
) -> st.SearchStrategy[list[BidsEntity]]:
    return st.lists(
        bids_entity(
            whitelist_entities=whitelist_entities,
            blacklist_entities=blacklist_entities,
        ),
        min_size=min_size,
        max_size=max_size,
        unique=True,
    ).filter(_filter_invalid_entity_lists)


@st.composite
def input_configs(
    draw: st.DrawFn, *, filters: bool = True, wildcards: bool = True, paths: bool = True
) -> InputConfig:
    sel_filters = (
        draw(
            st.one_of(
                st.dictionaries(
                    bids_entity().map(str),
                    st.one_of(
                        st.booleans(),
                        bids_value(),
                        st.lists(st.one_of(bids_value(), st.booleans())),
                    ),
                ),
                st.none(),
            )
        )
        if filters
        else None
    )

    sel_wildcards = (
        draw(st.one_of(st.lists(bids_entity().map(str)), st.none()))
        if wildcards
        else None
    )
    sel_path = (
        draw(
            st.one_of(
                st.text(
                    alphabet=st.characters(
                        blacklist_categories=("Cs",), blacklist_characters=("\x00",)
                    )
                ),
                st.none(),
            )
        )
        if paths
        else None
    )

    pybids_inputs: InputConfig = {}
    if sel_wildcards is not None:
        pybids_inputs.update({"wildcards": sel_wildcards})
    if sel_filters is not None:
        pybids_inputs.update({"filters": sel_filters})
    if sel_path is not None:
        pybids_inputs.update({"custom_path": sel_path})
    return pybids_inputs


def inputs_configs(
    *,
    keys: st.SearchStrategy[str] | None = None,
    min_size: int = 0,
    max_size: int | None = 5,
    filters: bool = True,
    wildcards: bool = True,
    paths: bool = True,
) -> st.SearchStrategy[InputsConfig]:
    return st.dictionaries(
        keys if keys is not None else st.text(min_size=1),
        input_configs(filters=filters, wildcards=wildcards, paths=paths),
        min_size=min_size,
        max_size=max_size,
    )


@st.composite
def zip_lists(
    draw: st.DrawFn,
    *,
    min_entities: int = 1,
    max_entities: int = 5,
    min_values: int = 1,
    max_values: int = 3,
    entities: list[BidsEntity] | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    restrict_patterns: bool = False,
    unique: bool = False,
    cull: bool = True,
) -> ZipList:
    # Generate multiple entity sets for different "file types"

    if entities is None:
        entities = draw(
            bids_entity_lists(
                min_size=min_entities,
                max_size=max_entities,
                blacklist_entities=blacklist_entities,
                whitelist_entities=whitelist_entities,
            )
        )

    def filter_ints(type_: str | None):
        def inner(s: str):
            if type_ == "int":
                return int(s) > 0 and not s.startswith("0")
            return True

        return inner

    values = {
        entity: draw(
            st.lists(
                bids_value(entity.match if restrict_patterns else ".*").filter(
                    filter_ints(entity.type if restrict_patterns else None)
                ),
                min_size=min_values,
                max_size=max_values,
                unique=(cull or unique),
            )
        )
        for entity in entities
    }

    combinations = list(it.product(*values.values()))
    used_combinations = (
        draw(
            st.lists(
                st.sampled_from(combinations),
                min_size=1,
                max_size=len(combinations),
                unique=unique,
            )
        )
        if cull and combinations
        else combinations
    )
    return helpers.get_zip_list(values, used_combinations)


@st.composite
def bids_component_row(
    draw: st.DrawFn,
    *,
    min_values: int = 1,
    max_values: int = 5,
    entity: BidsEntity | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    restrict_patterns: bool = False,
    unique: bool = False,
    path_safe: bool = False,
) -> BidsComponentRow:
    entity = entity or draw(
        bids_entity(
            blacklist_entities=blacklist_entities,
            whitelist_entities=whitelist_entities,
            path_safe=path_safe,
        )
    )
    values = draw(
        st.lists(
            bids_value(entity.match if restrict_patterns else ".*"),
            min_size=min_values,
            max_size=max_values,
            unique=unique,
        )
    )
    return BidsComponentRow(values, entity=entity.entity)


@st.composite
def bids_partial_components(
    draw: st.DrawFn,
    *,
    min_entities: int = 1,
    max_entities: int = 5,
    min_values: int = 1,
    max_values: int = 3,
    entities: list[BidsEntity] | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    restrict_patterns: bool = False,
    unique: bool = False,
    cull: bool = True,
    _allow_subclasses: bool = True,
) -> BidsPartialComponent:
    if _allow_subclasses:
        return draw(
            st.one_of(
                bids_components(
                    min_entities=min_entities,
                    max_entities=max_entities,
                    min_values=min_values,
                    max_values=max_values,
                    entities=entities,
                    blacklist_entities=blacklist_entities,
                    whitelist_entities=whitelist_entities,
                    restrict_patterns=restrict_patterns,
                    cull=cull,
                    unique=unique,
                ),
                bids_partial_components(
                    min_entities=min_entities,
                    max_entities=max_entities,
                    min_values=min_values,
                    max_values=max_values,
                    entities=entities,
                    blacklist_entities=blacklist_entities,
                    whitelist_entities=whitelist_entities,
                    restrict_patterns=restrict_patterns,
                    cull=cull,
                    unique=unique,
                    _allow_subclasses=False,
                ),
            )
        )
    zip_list = draw(
        zip_lists(
            min_entities=min_entities,
            max_entities=max_entities,
            min_values=min_values,
            max_values=max_values,
            entities=entities,
            blacklist_entities=blacklist_entities,
            whitelist_entities=whitelist_entities,
            restrict_patterns=restrict_patterns,
            cull=cull,
            unique=unique,
        )
    )

    return BidsPartialComponent(zip_lists=zip_list)


@st.composite
def bids_components(
    draw: st.DrawFn,
    *,
    min_entities: int = 1,
    max_entities: int = 5,
    min_values: int = 1,
    max_values: int = 3,
    entities: list[BidsEntity] | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    root: Path | None = None,
    name: str | None = None,
    restrict_patterns: bool = False,
    extra_entities: bool = True,
    blacklist_extra_entities: Container[BidsEntity | str] | None = None,
    whitelist_extra_entities: Container[BidsEntity | str] | None = None,
    unique: bool = False,
    cull: bool = True,
) -> BidsComponent:
    partial = draw(
        bids_partial_components(
            min_entities=min_entities,
            max_entities=max_entities,
            min_values=min_values,
            max_values=max_values,
            entities=entities,
            blacklist_entities=blacklist_entities,
            whitelist_entities=whitelist_entities,
            restrict_patterns=restrict_patterns,
            cull=cull,
            unique=unique,
            _allow_subclasses=False,
        )
    )
    path = draw(
        bids_path(
            root=root,
            entities=partial.zip_lists,
            extra_entities=extra_entities,
            blacklist_entities=blacklist_extra_entities,
            whitelist_entities=whitelist_extra_entities,
        )
    )
    return BidsComponent(
        name=name or draw(bids_value()),
        path=str(path),
        zip_lists=partial.zip_lists,
    )


@st.composite
def expandables(
    draw: st.DrawFn,
    *,
    max_entities: int = 5,
    min_values: int = 1,
    max_values: int = 3,
    entities: list[BidsEntity] | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    restrict_patterns: bool = False,
    unique: bool = False,
    cull: bool = True,
    path_safe: bool = False,
) -> Expandable:
    def get_entity(_entities: list[BidsEntity] | None) -> BidsEntity | None:
        if not _entities:
            return None
        return draw(st.sampled_from(_entities))

    return draw(
        st.one_of(
            bids_partial_components(
                min_entities=1,
                max_entities=max_entities,
                min_values=min_values,
                max_values=max_values,
                entities=entities,
                blacklist_entities=blacklist_entities,
                whitelist_entities=whitelist_entities,
                restrict_patterns=restrict_patterns,
                cull=cull,
                unique=unique,
            ),
            bids_component_row(
                min_values=min_values,
                max_values=max_values,
                entity=get_entity(entities),
                blacklist_entities=blacklist_entities,
                whitelist_entities=whitelist_entities,
                restrict_patterns=restrict_patterns,
                unique=unique,
                path_safe=path_safe,
            ),
        )
    )


@st.composite
def bids_input_lists(
    draw: st.DrawFn,
    *,
    min_size: int = 1,
    max_size: int = 5,
    entities: list[BidsEntity] | None = None,
) -> dict[str, list[str]]:
    # Generate multiple entity sets for different "file types"
    if entities is None:
        entities = draw(bids_entity_lists(min_size=min_size))

    return {
        entity.wildcard: draw(
            st.lists(bids_value(), min_size=min_size, max_size=max_size, unique=True)
        )
        for entity in entities
    }


@st.composite
def datasets(
    draw: st.DrawFn,
    *,
    root: Path | None = None,
    unique: bool = False,
    cull: bool = True,
) -> BidsDataset:
    ent1 = draw(bids_entity_lists(min_size=2, max_size=3))
    ent2 = copy.copy(ent1)
    extra = ent2.pop()
    assume(_filter_invalid_entity_lists(ent2))
    comp1 = draw(
        bids_components(
            entities=ent1, restrict_patterns=True, root=root, unique=unique, cull=cull
        )
    )
    comp2 = draw(
        bids_components(
            entities=ent2,
            restrict_patterns=True,
            root=root,
            blacklist_extra_entities=[extra],
            unique=unique,
            cull=cull,
        )
    )
    assume(comp1.input_name != comp2.input_name)
    return BidsDataset.from_iterable([comp1, comp2])


@st.composite
def datasets_one_comp(
    draw: st.DrawFn,
    *,
    root: Path | None = None,
    names: st.SearchStrategy[str] | None = None,
    blacklist_entities: Container[BidsEntity | str] | None = None,
    whitelist_entities: Container[BidsEntity | str] | None = None,
    unique: bool = False,
    cull: bool = True,
) -> BidsDataset:
    comp1 = draw(
        bids_components(
            min_entities=2,
            max_entities=3,
            restrict_patterns=True,
            root=root,
            name=draw(names) if names is not None else None,
            blacklist_entities=blacklist_entities,
            whitelist_entities=whitelist_entities,
            unique=unique,
            cull=cull,
        )
    )
    return BidsDataset.from_iterable([comp1])


@st.composite
def multiselect_dicts(
    draw: st.DrawFn,
    keys: st.SearchStrategy[_Ex_co],
    values: st.SearchStrategy[_T],
    *,
    min_size: int = 0,
    max_size: int | None = None,
) -> MultiSelectDict[_Ex_co, _T]:
    return MultiSelectDict(
        draw(
            st.dictionaries(
                keys,
                values,
                min_size=min_size,
                max_size=max_size,
            )
        )
    )


if TYPE_CHECKING:

    def everything() -> st.SearchStrategy[Any]: ...

    def everything_except(*excluded_types: type[Any]) -> st.SearchStrategy[Any]: ...

else:

    def everything() -> st.SearchStrategy[Any]:
        return st.from_type(type).flatmap(st.from_type)

    def everything_except(*excluded_types: type[Any]) -> st.SearchStrategy[Any]:
        return (
            st.from_type(type)
            .flatmap(st.from_type)
            .filter(lambda s: not isinstance(s, excluded_types))
        )


def _is_hashable(item: Any, /):
    try:
        hash(item)
    except TypeError:
        return False
    return True


def _supports_eq(item: Any, /):
    try:
        item == 0  # type: ignore  # noqa: B015
    except Exception:  # noqa: BLE001
        return False
    return True


def hashables() -> st.SearchStrategy[Hashable]:
    return everything().filter(_is_hashable)


def partially_ordered() -> st.SearchStrategy[Any]:
    return everything().filter(_supports_eq)



from __future__ import annotations

import re
import sys
from collections.abc import Container, Iterable, Iterator, Sequence
from typing import (
    Any,
    AnyStr,
    Generic,
    SupportsIndex,
    TypeVar,
    overload,
)

from typing_extensions import Self, override

_T = TypeVar("_T")
_K = TypeVar("_K", bound="str")
_T_co = TypeVar("_T_co", covariant=True)


class ImmutableList(Sequence[_T_co], Generic[_T_co]):
    """Subclassable tuple equivalent.

    Mimics a tuple in every way, but readily supports subclassing. Data is stored on a
    private attribute ``_data``. Subclasses must not override this attribute. To avoid
    accidental modification, subclasses should avoid interacting with ``_data``, using
    the relevant ``super()`` calls to access internal data instead (e.g. use
    ``super().__getitem__(index)`` rather than ``self._data[index]``).

    Unlike tuples, only a single type parameter is supported. In other words,
    ``ImmutableList`` cannot be specified via type hints as a fixed length sequence
    containing heterogeneous items. A tuple specified as ``tuple[str, int, str]`` would
    be specified as ``ImmutableList[str | int]``
    """

    def __init__(self, iterable: Iterable[_T_co] = (), /):
        self._data = tuple(iterable)

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({list(self._data)})"

    @override
    def __contains__(self, value: object) -> bool:
        return value in self._data

    @override
    def __hash__(self):
        return hash(self._data)

    @override
    def __iter__(self) -> Iterator[_T_co]:
        return iter(self._data)

    @override
    def __reversed__(self) -> Iterator[_T_co]:
        return reversed(self._data)

    @override
    def __len__(self) -> int:
        return len(self._data)

    def __bool__(self) -> bool:
        return bool(self._data)

    @overload
    def __getitem__(self, index: int) -> _T_co: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    @override
    def __getitem__(self, index: int | slice) -> _T_co | Self:
        if isinstance(index, slice):
            return self.__class__(self._data[index])
        return self._data[index]

    def __lt__(self, value: tuple[_T_co, ...] | Self, /) -> bool:
        if isinstance(value, ImmutableList):
            return self._data < value._data
        return self._data < value

    def __le__(self, value: tuple[_T_co, ...] | Self, /) -> bool:
        if isinstance(value, ImmutableList):
            return self._data <= value._data
        return self._data <= value

    def __gt__(self, value: tuple[_T_co, ...] | Self, /) -> bool:
        if isinstance(value, ImmutableList):
            return self._data > value._data
        return self._data > value

    def __ge__(self, value: tuple[_T_co, ...] | Self, /) -> bool:
        if isinstance(value, ImmutableList):
            return self._data >= value._data
        return self._data >= value

    @override
    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, tuple):
            return self._data == value
        if isinstance(value, ImmutableList):
            return self._data == value._data  # type: ignore
        return False

    def __add__(self, value: tuple[_T_co, ...] | ImmutableList[_T_co], /) -> Self:
        if isinstance(value, ImmutableList):
            return self.__class__(self._data + value._data)
        return self.__class__(self._data + value)

    def __mul__(self, value: SupportsIndex, /) -> Self:
        return self.__class__(self._data * value)

    def __rmul__(self, value: SupportsIndex, /) -> Self:
        return self.__class__(value * self._data)

    @override
    def count(self, value: Any) -> int:
        return self._data.count(value)

    @override
    def index(
        self, value: Any, start: SupportsIndex = 0, stop: SupportsIndex = sys.maxsize
    ) -> int:
        return self._data.index(value, start, stop)


class RegexContainer(Generic[AnyStr], Container[AnyStr]):
    """Container that tests if a string matches a regex using the ``in`` operator.

    Constructed with a regex expression. Supports inclusion tests for strings using
    ``in``. Strings matching the regex (using ``re.match``) will return ``True``
    """

    def __init__(self, template: AnyStr | re.Pattern[AnyStr]):
        self.template: re.Pattern[AnyStr] = re.compile(template)

    def __contains__(self, x: object, /):
        if isinstance(x, type(self.template.pattern)):
            return self.template.match(x) is not None
        return False


class ContainerBag(Container[_T]):
    """Container to hold other containers.

    Useful because list(Container) isn't guaranteed to work, so this lets us merge
    Containers in a type safe way.
    """

    def __init__(self, *entries: Container[_T]):
        self.entries = entries

    def __contains__(self, x: object, /) -> bool:
        return any(x in entry for entry in self.entries)


class MultiSelectDict(dict[_K, _T]):
    """Dict supporting selection of multiple keys using tuples.

    If a single key is given, the item associated with that key is returned just as in a
    regular dict. If multiple, comma-seperated keys are given, (e.g. a tuple), a new
    ``MultiSelectDict`` will be returned containing the keys given and their values::

        >>> mydict = MultiSelectDict({
        ...     "foo": "bar",
        ...     "hello": "world",
        ...     "fee": "fie",
        ... })
        >>> mydict["foo"]
        'bar'
        >>> mydict["foo", "hello"]
        {'foo': 'bar', 'hello': 'world'}

    The new ``MultiSelectDict`` is a "view" of the original data. Any mutations made to
    the values will be reflected in the original object::

        >>> mydict = MultiSelectDict({
        ...     "foo": [1, 2, 3, 4],
        ...     "hello": "world",
        ...     "fee": "fie",
        ... })
        >>> view = mydict["foo", "hello"]
        >>> view["foo"].append(5)
        >>> mydict
        {'foo': [1, 2, 3, 4, 5], 'hello': 'world', 'fee': 'fie'}

    The keys of the new ``MultiSelectDict`` will be inserted in the order provided in
    the selector::

        >>> mydict = MultiSelectDict({
        ...     "foo": "bar",
        ...     "hello": "world",
        ...     "fee": "fie",
        ... })
        >>> mydict["hello", "foo"]
        {'hello': 'world', 'foo': 'bar'}

    Tuples of length 1 will still return a ``MultiSelectDict``::

        >>> mydict = MultiSelectDict({
        ...     "foo": [1, 2, 3, 4],
        ...     "hello": "world",
        ...     "fee": "fie",
        ... })
        >>> mydict["foo",]
        {'foo': [1, 2, 3, 4]}

    If a key is given multiple times, the extra instances of the key will be ignored::

        >>> mydict = MultiSelectDict({
        ...     "foo": [1, 2, 3, 4],
        ...     "hello": "world",
        ...     "fee": "fie",
        ... })
        >>> mydict["foo", "foo", "foo"]
        {'foo': [1, 2, 3, 4]}
    """

    @overload
    def __getitem__(self, key: _K, /) -> _T: ...

    @overload
    def __getitem__(self, key: tuple[_K, ...], /) -> Self: ...

    def __getitem__(self, key: _K | tuple[_K, ...], /) -> _T | Self:
        if isinstance(key, tuple):
            # Use dict.fromkeys for de-duplication
            return self.__class__({key: self[key] for key in dict.fromkeys(key)})
        return super().__getitem__(key)



"""
This type stub file was generated by pyright.
"""

from __future__ import annotations

from typing import Any

from ruamel.yaml.main import *

from .cyaml import *

if False: ...
_package_data: dict[Any, Any] = ...
version_info = ...
__version__ = ...
__with_libyaml__ = ...



from __future__ import annotations

import re
import subprocess as sp
import sys
from pathlib import Path

import jinja2.parser
from jinja2 import nodes
from jinja2.ext import Extension
from typing_extensions import override


class GitConfigExtension(Extension):
    """Retrieve settings from git configuration.

    Shortcode activated with ``gitconfig`` plus the key name::

        {% gitconfig "user.name" %}
    """

    tags = {"gitconfig"}  # noqa: RUF012
    _config: dict[str, str]

    def __init__(self, env: jinja2.Environment) -> None:
        self._config = {}

        try:
            config_list = sp.check_output(
                [executable(), "config", "-l"], stderr=sp.STDOUT
            ).decode()

            m = re.findall("(?ms)^([^=]+)=(.*?)$", config_list)
            if m:
                for group in m:
                    self._config[group[0]] = group[1]
        except (sp.CalledProcessError, OSError):
            pass

    def get(self, key: str, default: str | None = None) -> str | None:
        """Get specified configuration key."""
        return self._config.get(key, default)

    def __getitem__(self, item: str) -> str:
        return self._config[item]

    @override
    def parse(self, parser: jinja2.parser.Parser):
        lineno = next(parser.stream).lineno

        node = parser.parse_expression()

        if not isinstance(node, nodes.Const):
            msg = "Argument to `gitconfig` must be a string"
            raise TypeError(msg)
        call_method = self.call_method(
            "get",
            [node],
            lineno=lineno,
        )
        return nodes.Output([call_method], lineno=lineno)


def executable() -> str:
    """Retrieve os-dependent command for git."""
    _executable = None

    if sys.platform == "win32":
        # Finding git via where.exe
        where = "%WINDIR%\\System32\\where.exe"
        paths = sp.check_output([where, "git"], encoding="oem").split("\n")
        for path in paths:
            if not path:
                continue

            _path = Path(path.strip())
            try:
                _path.relative_to(Path.cwd())
            except ValueError:
                _executable = str(_path)

                break
    else:
        _executable = "git"

    if _executable is None:  # type: ignore
        msg = "Unable to find a valid git executable"
        raise RuntimeError(msg)

    return _executable



from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Any

import pytest
from hypothesis import given
from pytest_mock import MockerFixture

from snakebids import bidsapp
from snakebids.app import SnakeBidsApp, sb_plugins
from snakebids.cli import add_dynamic_args
from tests import strategies as sb_st


class TestDeprecations:
    def test_parser_deprecated(self):
        app = SnakeBidsApp(Path(), parser=argparse.ArgumentParser())
        with pytest.warns(UserWarning, match="`SnakeBidsApp.parser` is deprecated"):
            app._check_deprecations()

    def test_config_deprecated(self):
        app = SnakeBidsApp(Path(), config={})
        with pytest.warns(UserWarning, match="`SnakeBidsApp.config` is deprecated"):
            app._check_deprecations()

    def test_args_deprecated(self):
        app = SnakeBidsApp(Path(), args=[])
        with pytest.warns(UserWarning, match="`SnakeBidsApp.args` is deprecated"):
            app._check_deprecations()

    def test_version_deprecated(self):
        app = SnakeBidsApp(Path(), version="foo")
        with pytest.warns(UserWarning, match="`SnakeBidsApp.version` is deprecated"):
            app._check_deprecations()


def test_arguments_carried_forward(mocker: MockerFixture):
    mocker.patch.object(bidsapp.run, "_Runner")
    snakemake = mocker.spy(sb_plugins, "SnakemakeBidsApp")
    SnakeBidsApp(
        Path(),
        configfile_path=Path("config"),
        snakefile_path=Path("Snakefile"),
    ).run_snakemake()
    snakemake.assert_called_once_with(
        snakemake_dir=Path(),
        configfile_path=Path("config"),
        snakefile_path=Path("Snakefile"),
    )


def test_str_converted_to_path(mocker: MockerFixture):
    mocker.patch.object(bidsapp.run, "_Runner")
    snakemake = mocker.spy(sb_plugins, "SnakemakeBidsApp")
    SnakeBidsApp(
        "",
        configfile_path="config",
        snakefile_path="Snakefile",
    ).run_snakemake()
    snakemake.assert_called_once_with(
        snakemake_dir=Path(),
        configfile_path=Path("config"),
        snakefile_path=Path("Snakefile"),
    )


def test_plugins_carried_forward(mocker: MockerFixture):
    mocker.stopall()
    mocker.patch.object(bidsapp.run, "_Runner")
    app_spy = mocker.spy(bidsapp, "app")
    SnakeBidsApp(
        Path(),
        configfile_path=Path("config"),
        snakefile_path=Path("Snakefile"),
        plugins=["one", "two"],  # type: ignore
    ).run_snakemake()
    app_spy.assert_called_once()
    plugins = app_spy.call_args[0][0]
    assert isinstance(plugins, list)
    assert len(plugins) == 3  # type: ignore  # noqa: PLR2004


def test_parser_can_be_directly_accessed(tmpdir: Path):
    config_path = Path(tmpdir) / "config.json"
    config_path.write_text("{}")
    app = SnakeBidsApp(
        Path(),
        configfile_path=config_path,
        snakefile_path=Path("Snakefile"),
    )
    assert app.parser is app._app.parser


def test_config_can_be_directly_accessed(tmpdir: Path):
    config_path = Path(tmpdir) / "config.json"
    config_path.write_text("{}")
    app = SnakeBidsApp(
        Path(),
        configfile_path=config_path,
        snakefile_path=Path("Snakefile"),
    )
    assert app.config is app._app.config


@given(
    parser=sb_st.everything(),
    parse_args=sb_st.everything(),
    pybids_inputs=sb_st.everything(),
)
def test_add_dynamic_args_raises_warning(
    parser: Any, parse_args: Any, pybids_inputs: Any
):
    with pytest.warns(UserWarning, match="is deprecated and no longer has any effect"):
        add_dynamic_args(parser, parse_args, pybids_inputs)


class TestGenBoutiques:
    def test_boutiques_descriptor(self, tmp_path: Path):
        configpth = tmp_path / "config.json"
        configpth.write_text("{}")
        app = SnakeBidsApp(
            Path(),
            configfile_path=configpth,
            snakefile_path=Path("Snakefile"),
            plugins=["one", "two"],  # type: ignore
        )
        descriptor_path = tmp_path / "descriptor.json"
        app.create_descriptor(descriptor_path)
        with open(descriptor_path, encoding="utf-8") as descriptor_file:
            descriptor_json = json.load(descriptor_file)
            assert "command-line" in descriptor_json
            assert "inputs" in descriptor_json



from __future__ import annotations

import functools as ft
import json
import os
import re
import textwrap
from collections.abc import Callable, Iterable, Sequence
from os import PathLike
from pathlib import Path
from typing import Any, Protocol, TypeAlias, TypeVar

import attrs
import importlib_resources as impr
import more_itertools as itx
from typing_extensions import NotRequired, TypedDict

from snakebids import resources, types
from snakebids.utils.user_property import UserProperty

_T = TypeVar("_T")


class BidsTag(TypedDict):
    """Interface for BidsTag configuration."""

    tag: str
    before: NotRequired[str]
    match: str
    after: NotRequired[str]
    leader: NotRequired[bool]


BidsTags: TypeAlias = "dict[str, BidsTag]"


DEPRECATION_FLAG = "<!DEPRECATED!>"
"""Sentinel string to mark deprecated config features"""


@ft.lru_cache
def read_bids_tags(bids_json: Path | None = None) -> BidsTags:
    """Read the bids tags we are aware of from a JSON file.

    This is used specifically for compatibility with pybids, since some tag keys
    are different from how they appear in the file name, e.g. ``subject`` for
    ``sub``, and ``acquisition`` for ``acq``.

    Parameters
    ----------
    bids_json
        Path to JSON file to use, if not specified will use
        ``bids_tags.json`` in the snakebids module.

    Returns
    -------
    dict:
        Dictionary of bids tags
    """
    if bids_json:
        with bids_json.open("r") as infile:
            return json.load(infile)
    return json.loads(impr.files(resources).joinpath("bids_tags.json").read_text())


@attrs.frozen(hash=False)
class BidsEntity:
    """Bids entities with tag and wildcard representations."""

    entity: str = attrs.field(converter=str)

    def __str__(self) -> str:
        return self.entity

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return self.entity == other
        if isinstance(other, BidsEntity):
            return self.entity == other.entity
        return False

    def __hash__(self) -> int:
        return hash(self.entity)

    def __lt__(self, other: BidsEntity | str):
        if isinstance(other, str):
            return self.entity < other
        return self.entity < other.entity

    @property
    def tag(self) -> str:
        """Get the bids tag version of the entity.

        For entities in the bids spec, the tag is the short version of the entity
        name. Otherwise, the tag is equal to the entity.
        """
        tags = read_bids_tags()
        return (
            tags[self.entity]["tag"]
            if self.entity in tags and "tag" in tags[self.entity]
            else self.entity
        )

    @property
    def match(self) -> str:
        """Get regex of acceptable value matches.

        If no pattern is associated with the entity, the default pattern is a word with
        letters and numbers
        """
        tags = read_bids_tags()
        return (
            tags[self.entity]["match"]
            if self.entity in tags and "match" in tags[self.entity]
            else "[a-zA-Z0-9]+"
        )

    @property
    def before(self) -> str:
        """Regex str to search before value in paths."""
        tags = read_bids_tags()
        # Need to explicitly annotate the default here and in .after because tags is a
        # dict of `BidsTag`, a `TypedDict`. So putting unannotated dicts directly as
        # default leads to a type error
        _def: dict[Any, Any] = {}
        return tags.get(self.entity, _def).get("before", f"{self.tag}-")

    @property
    def after(self) -> str:
        """Regex str to search after value in paths."""
        tags = read_bids_tags()
        # See note in .before
        _def: dict[Any, Any] = {}
        return tags.get(self.entity, _def).get("after", "")

    @property
    def regex(self) -> re.Pattern[str]:
        """Complete pattern to match when searching in paths.

        Contains three capture groups, the first corresponding to "before", the second
        to "value", and the third to "after"
        """
        return re.compile(f"({self.before})({self.match})({self.after})")

    @property
    def wildcard(self) -> str:
        """Get the snakebids {wildcard}.

        The wildcard is generally equal to the tag, i.e. the short version of the entity
        name, except for subject and session, which use the full name name. This is to
        ensure compatibility with the bids function
        """
        # HACK FIX FOR acq vs acquisition etc -- should
        # eventually update the bids() function to also use
        # bids_tags.json, where e.g. acquisition -> acq is
        # defined.. -- then, can use wildcard_name instead
        # of out_name..
        if self.entity in ["subject", "session"]:
            return self.entity
        return self.tag

    @property
    def type(self) -> str | None:
        """Get the type of the entity.

        Returns None if type unspecified.
        """
        tags = read_bids_tags()
        # See note in .before
        _def: dict[Any, Any] = {}
        return tags.get(self.entity, _def).get("type")

    @classmethod
    def from_tag(cls, tag: str) -> BidsEntity:
        """Return the entity associated with the given tag, if found.

        If not associated entity is found, the tag itself is used as the entity name

        Parameters
        ----------
        tag : str
            tag to search

        Returns
        -------
        BidsEntity
        """
        for entity, props in read_bids_tags().items():
            if props.get("tag", None) == tag:
                return cls(entity)
        return cls(tag)

    @classmethod
    def normalize(cls, item: str | BidsEntity, /) -> BidsEntity:
        """Return the entity associated with the given item, if found.

        Supports both strings and BidsEntities as input. Unlike the constructor, if a
        tag name is given, the associated entity will be returned. If no associated
        entity is found, the tag itself is used as the entity name

        Parameters
        ----------
        item
            tag to search
        """
        if isinstance(item, BidsEntity):
            return item
        return cls.from_tag(item)


def matches_any(
    item: _T,
    match_list: Iterable[_T],
    match_func: types.BinaryOperator[_T, object],
    *args: Any,
) -> bool:
    """Test if item matches any of the items in match_list.

    Parameters
    ----------
    item
        Item to test
    match_list
        Items to compare with
    match_func
        Function to test equality. Defaults to basic equality (``==``) check
    """
    return any(match_func(match, item, *args) for match in match_list)


class BidsParseError(Exception):
    """Exception raised for errors encountered in the parsing of Bids paths."""

    def __init__(self, path: str, entity: BidsEntity) -> None:
        self.path = path
        self.entity = entity
        super().__init__(path, entity)


class _Documented(Protocol):
    __doc__: str


def property_alias(
    prop: _Documented,
    label: str | None = None,
    ref: str | None = None,
    copy_extended_docstring: bool = False,
) -> Callable[[Callable[[Any], _T]], UserProperty[_T]]:
    """Set property as an alias for another property.

    Copies the docstring from the aliased property to the alias

    Parameters
    ----------
    prop : property
        Property to alias
    label
        Text to use in link to aliased property
    ref
        Name of the property to alias
    copy_extended_docstring
        If True, copies over the entire docstring, in addition to the summary line

    Returns
    -------
    property
    """

    def inner(func: Callable[[Any], _T], /) -> UserProperty[_T]:
        alias = UserProperty(func)
        if label:
            link = f":attr:`{label} <{ref}>`" if ref else label
        else:
            link = f":attr:`{ref}`" if ref else None
        labeltxt = f"Alias of {link}\n\n" if link else ""
        if copy_extended_docstring:
            alias.__doc__ = f"{labeltxt}{prop.__doc__}"
        else:
            alias.__doc__ = f"{labeltxt}{prop.__doc__.splitlines()[0]}"
        return alias

    return inner


def surround(s: Iterable[str] | str, object_: str, /) -> Iterable[str]:
    """Surround a string or each string in an iterable with characters."""
    for item in itx.always_iterable(s):
        yield object_ + item + object_


def _hash_cols(tbl: Iterable[Sequence[str]]):
    return {hash(col) for col in zip(*tbl, strict=True)}


def zip_list_eq(first: types.ZipListLike, second: types.ZipListLike, /):
    """Compare two zip lists, allowing the order of columns to be irrelevant."""
    if not first and not second:
        return True

    if set(first) != set(second):
        return False

    return _hash_cols(first.values()) == _hash_cols(second[key] for key in first)


def get_first_dir(path: str) -> str:
    """Return the top level directory in a path.

    If absolute, return the root. This function is necessary to handle paths with
    ``./``, as ``pathlib.Path`` filters this out.
    """
    if os.path.isabs(path):
        return Path(path).root
    parent, child = os.path.split(path)
    if parent:
        return get_first_dir(parent)
    return child


def to_resolved_path(path: str | PathLike[str]):
    """Convert provided object into resolved path."""
    return Path(path).resolve()


def get_wildcard_dict(entities: str | Iterable[str], /) -> dict[str, str]:
    """Turn entity strings into wildcard dicts as {"entity": "{entity}"}."""
    return {entity: f"{{{entity}}}" for entity in itx.always_iterable(entities)}


def entity_to_wildcard(entities: str | Iterable[str], /):
    """Turn entity strings into wildcard dicts as {"entity": "{entity}"}."""
    return {entity: f"{{{entity}}}" for entity in itx.always_iterable(entities)}


def text_fold(text: str):
    """Fold a block of text into a single line as in yaml folded multiline string."""
    return " ".join(textwrap.dedent(text).strip().splitlines())



from __future__ import annotations

import argparse
import copy
import tempfile
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import pytest
from hypothesis import given
from hypothesis import strategies as st
from pytest_mock import MockerFixture
from pytest_mock.plugin import MockType

import snakebids.plugins.snakemake as sn_app
import snakebids.snakemake_compat
from snakebids import bidsapp
from snakebids.exceptions import ConfigError, RunError
from snakebids.plugins import Version
from snakebids.utils.utils import DEPRECATION_FLAG
from tests import strategies as sb_st
from tests.helpers import allow_function_scoped


class TestFindSnakefileConfig:
    @allow_function_scoped
    @given(snakefile=st.sampled_from(sn_app.SNAKEFILE_CHOICES))
    def test_finds_snakefile(self, tmpdir: Path, snakefile: str):
        tmp = Path(tempfile.mkdtemp(dir=tmpdir))
        snakefile_path = tmp / snakefile
        snakefile_path.parent.mkdir(parents=True, exist_ok=True)
        snakefile_path.touch()
        app = sn_app.SnakemakeBidsApp(snakemake_dir=tmp, configfile_path=None)
        assert app.snakefile_path == snakefile_path

    def test_errors_if_no_snakefile_found(self, tmpdir: Path):
        tmp = Path(tempfile.mkdtemp(dir=tmpdir))
        with pytest.raises(ConfigError, match="Error: no Snakefile"):
            sn_app.SnakemakeBidsApp(snakemake_dir=tmp, configfile_path=None)

    @allow_function_scoped
    @given(configfile=st.sampled_from(sn_app.CONFIGFILE_CHOICES))
    def test_finds_config_file(self, tmpdir: Path, configfile: str):
        tmp = Path(tempfile.mkdtemp(dir=tmpdir))
        configfile_path = tmp / configfile
        configfile_path.parent.mkdir(parents=True, exist_ok=True)
        configfile_path.touch()
        app = sn_app.SnakemakeBidsApp(snakemake_dir=tmp, snakefile_path=Path())
        assert app.configfile_path == configfile_path

    def test_errors_if_no_configfile_found(self, tmpdir: Path):
        tmp = Path(tempfile.mkdtemp(dir=tmpdir))
        with pytest.raises(ConfigError, match="Error: no config file"):
            sn_app.SnakemakeBidsApp(snakemake_dir=tmp, snakefile_path=Path())


class TestAddCliArguments:
    def test_snakemake_help_arg_added(self, mocker: MockerFixture):
        mock = mocker.patch.object(sn_app, "snakemake_main")
        smk = sn_app.SnakemakeBidsApp.create_empty()
        parser = argparse.ArgumentParser()
        smk.add_cli_arguments(parser)
        parser.parse_args(["--help-snakemake"])
        mock.assert_called_once_with(["-h"])


class TestResolvePath:
    @pytest.fixture
    def arg_dict(self) -> dict[str, str | list[str]]:
        return {
            "bids_dir": "path/to/input",
            "output_dir": "path/to/output",
            "analysis_level": "participant",
            "--derivatives": ["path/to/deriv1", "path/to/deriv2"],
        }

    def test_does_not_change_dict_without_paths(
        self, arg_dict: Mapping[str, str | Sequence[str]]
    ):
        arg_dict_copy = copy.deepcopy(arg_dict)
        resolved = {key: sn_app._resolve_path(value) for key, value in arg_dict.items()}
        assert resolved == arg_dict_copy

    def test_resolves_all_paths(
        self, arg_dict: dict[str, str | Path | Sequence[str | Path]]
    ):
        derivative_paths = [Path("path/to/deriv1"), Path("path/to/deriv2")]
        arg_dict["--derivatives"] = derivative_paths
        arg_dict_copy = copy.deepcopy(arg_dict)
        arg_dict_copy["--derivatives"] = [p.resolve() for p in derivative_paths]
        resolved = {key: sn_app._resolve_path(value) for key, value in arg_dict.items()}
        assert resolved == arg_dict_copy

    def test_filter_dict(self, arg_dict: dict[str, dict[str, str]]):
        filter_dict = {"test_key": "test_value"}
        arg_dict["filter_test"] = filter_dict
        arg_dict_copy = copy.deepcopy(arg_dict)
        resolved = {key: sn_app._resolve_path(value) for key, value in arg_dict.items()}
        assert resolved == arg_dict_copy


class TestUpdateNamespace:
    @given(
        config=st.dictionaries(
            st.text(),
            st.text(
                st.characters(
                    blacklist_categories=["Cs"], blacklist_characters=["\x00"]
                )
            ).map(Path),
        )
    )
    def test_resolves_paths(self, config: dict[str, Path]):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        namespace = {
            **config,
            "output_dir": "foo",
            "bids_dir": "foo",
            "some_key": "foo",
        }
        smk.update_cli_namespace(namespace, {})
        for key, path in namespace.items():
            if key == "some_key":
                assert path == "foo"
            else:
                assert isinstance(path, Path)
                assert path.is_absolute()

    @given(targets=st.dictionaries(st.text(), st.text(), min_size=1), data=st.data())
    def test_target_gets_set(self, targets: dict[str, str], data: st.DataObject):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        level = data.draw(st.sampled_from(list(targets)))
        namespace = {"analysis_level": level}
        config = {"targets_by_analysis_level": targets}
        smk.update_cli_namespace(namespace, config)
        assert config["snakemake_target"] == targets[level]

    @given(value=st.sampled_from([True, False]))
    def test_force_output_is_set(self, value: bool):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        namespace = {"force_output": value}
        smk.update_cli_namespace(namespace, {})
        assert smk.force_output == value
        assert "force_output" not in namespace


def get_io_mocks(mocker: MockerFixture):
    mocker.stopall()
    mocker.patch.object(sn_app.impm, "version", return_value="version")
    mocker.patch.object(
        snakebids.snakemake_compat.configfile,
        "open",
        mocker.mock_open(read_data='{"a_key": "a value"}'),  # pyright: ignore[reportUnknownMemberType]
    )
    return {
        "write_output_mode": mocker.patch.object(sn_app, "write_output_mode"),
        "prepare_output": mocker.patch.object(sn_app, "prepare_bidsapp_output"),
        "write_config": mocker.patch.object(sn_app, "write_config"),
        "snakemake": mocker.patch.object(sn_app, "snakemake_main"),
    }


class TestFinalizeConfig:
    def check_write_config(
        self,
        mock: MockType,
        config_file: Path,
        config: dict[str, Any],
        root: str | Path,
    ):
        mock.assert_called_once_with(
            config_file=config_file,
            data=dict(
                config,
                snakemake_version="version",
                snakebids_version="version",
                root=root,
                app_version="unknown",
                snakemake_dir=Path().resolve(),
                snakefile=Path("Snakefile"),
            ),
            force_overwrite=True,
        )

    @given(
        path=sb_st.paths(resolve=True).filter(
            lambda p: p != Path().resolve()
            and not str(p).startswith(str(Path("results").resolve()))
        )
    )
    @allow_function_scoped
    def test_normally_preserves_paths(
        self,
        mocker: MockerFixture,
        path: Path,
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        io_mocks = get_io_mocks(mocker)

        config = {"output_dir": path}

        smk.finalize_config(config)

        assert config["output_dir"] == path
        assert smk.cwd == path
        io_mocks["write_output_mode"].assert_not_called()
        io_mocks["prepare_output"].assert_called_once_with(path, False)

        new_config = path.resolve() / "snakebids.yaml"
        self.check_write_config(
            io_mocks["write_config"], config_file=new_config, config=config, root=""
        )

    @pytest.mark.parametrize("path", [Path().resolve(), Path("results").resolve()])
    def test_output_in_app_triggers_workflow_mode(
        self,
        mocker: MockerFixture,
        path: Path,
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        io_mocks = get_io_mocks(mocker)

        config = {"output_dir": path}

        smk.finalize_config(config)

        assert config["output_dir"] == Path("results").resolve()
        assert smk.cwd == Path().resolve()
        io_mocks["write_output_mode"].assert_called_once_with(
            Path(".snakebids").resolve(), "workflow"
        )
        io_mocks["prepare_output"].assert_not_called()

        new_config = Path("snakebids.yaml").resolve()
        self.check_write_config(
            io_mocks["write_config"],
            config_file=new_config,
            config=config,
            root=Path("results"),
        )

    @given(tail=sb_st.paths(absolute=False))
    @allow_function_scoped
    def test_output_under_results_triggers_workflow_mode(
        self,
        mocker: MockerFixture,
        tail: Path,
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        io_mocks = get_io_mocks(mocker)
        path = Path("results", tail).resolve()

        config = {"output_dir": path}

        smk.finalize_config(config)

        assert config["output_dir"] == path
        assert smk.cwd == Path().resolve()
        io_mocks["write_output_mode"].assert_called_once_with(
            Path(".snakebids").resolve(), "workflow"
        )
        io_mocks["prepare_output"].assert_not_called()

        new_config = Path("snakebids.yaml").resolve()
        self.check_write_config(
            io_mocks["write_config"],
            config_file=new_config,
            config=config,
            root=Path("results", tail),
        )

    @given(
        output=st.sampled_from(["", "results"]).map(Path) | sb_st.paths(resolve=True),
        configfile=sb_st.paths(absolute=False),
    )
    @allow_function_scoped
    def test_relative_configfile_gets_transferred(
        self, mocker: MockerFixture, output: Path, configfile: Path
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        smk.configfile_path = configfile
        get_io_mocks(mocker)
        config = {"output_dir": output}

        smk.finalize_config(config)

        assert smk.configfile_outpath == smk.cwd / configfile

    @given(
        output=st.sampled_from(["", "results"]).map(Path) | sb_st.paths(absolute=False),
        configfile=sb_st.paths(absolute=True),
    )
    @allow_function_scoped
    def test_non_relative_configfile_is_preserved(
        self, mocker: MockerFixture, output: Path, configfile: Path
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        smk.configfile_path = configfile
        get_io_mocks(mocker)
        config = {"output_dir": output}

        smk.finalize_config(config)

        assert smk.configfile_outpath == configfile

    @given(
        output=st.sampled_from(["", "results"]).map(Path) | sb_st.paths(resolve=True),
        configfile=sb_st.paths(),
        config_output=sb_st.paths(),
    )
    @allow_function_scoped
    def test_specified_configfile_output_always_preserved(
        self, mocker: MockerFixture, output: Path, configfile: Path, config_output: Path
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        smk.configfile_path = configfile
        smk.configfile_outpath = config_output
        get_io_mocks(mocker)
        config = {"output_dir": output}

        smk.finalize_config(config)

        assert smk.configfile_outpath == config_output

    @given(
        err=st.text(), output=sb_st.paths(absolute=True, min_segments=1, max_segments=1)
    )
    @allow_function_scoped
    def test_exits_on_run_error(
        self,
        mocker: MockerFixture,
        capsys: pytest.CaptureFixture[str],
        err: str,
        output: Path,
    ):
        smk = sn_app.SnakemakeBidsApp.create_empty()
        mocks = get_io_mocks(mocker)
        mocks["prepare_output"].side_effect = RunError(err)
        config = {"output_dir": output.resolve()}

        with pytest.raises(SystemExit):
            smk.finalize_config(config)
        cap = capsys.readouterr()
        assert err == cap.err[:-1]


class TestVersion:
    def test_get_app_version_no_package(
        self, mocker: MockerFixture, caplog: pytest.LogCaptureFixture
    ):
        plugin = sn_app.SnakemakeBidsApp(
            snakemake_dir=Path(),
            configfile_path=None,
            snakefile_path=Path("Snakefile"),
        )
        assert caplog.text == ""
        get_io_mocks(mocker)
        plugin.finalize_config(
            {"output_dir": Path("foo"), "plugins.version.version": "unknown"}
        )
        assert "App version could not be identified" in caplog.text

    def test_missing_version_changed_to_unknown(self, mocker: MockerFixture):
        plugin = sn_app.SnakemakeBidsApp(
            snakemake_dir=Path(),
            configfile_path=None,
            snakefile_path=Path("Snakefile"),
        )
        mocks = get_io_mocks(mocker)
        plugin.finalize_config({"output_dir": Path("foo")})

        mocks["write_config"].assert_called_once_with(
            config_file=Path("foo/snakebids.yaml"),
            data={
                "output_dir": Path("foo"),
                "snakemake_version": "version",
                "snakebids_version": "version",
                "root": "",
                "app_version": "unknown",
                "snakemake_dir": Path().resolve(),
                "snakefile": Path("Snakefile"),
            },
            force_overwrite=True,
        )


@pytest.mark.parametrize("root", ["app", "app/results", "foo/bar/lo", "/absolute/path"])
def test_integration(
    mocker: MockerFixture,
    root: str,
):
    # Get mocks for all the io functions
    io_mocks = get_io_mocks(mocker)

    # Prepare app and initial config values
    plugin = sn_app.SnakemakeBidsApp(
        Path("app"),
        snakefile_path=Path("app/Snakefile"),
        configfile_path=Path("app/config/config.yaml"),
    )
    # dummy db_path: all io functions are mocked, so this can be arbitrary str
    db_path = "/path/to/db"
    app = bidsapp.app(
        plugins=[plugin, Version(distribution="snakebidstesting")],
    )
    app.config["targets_by_analysis_level"] = {"participant": "all"}
    app.config["snakemake_args"] = []
    app.config["pybids_inputs"] = {
        "bold": {
            "filters": {"suffix": "bold", "extension": ".nii.gz", "datatype": "func"},
            "wildcards": ["subject", "session", "acquisition", "task", "run"],
        }
    }
    app.config["parse_args"] = {
        "--additional-arg": {"nargs": "+"},
        "--derivatives": {"help": "overwrite this arg", "nargs": "*"},
    }

    # Prepare expected config
    expected_config = copy.deepcopy(app.config)
    expected_config.update(
        {
            "root": "",
            "a_key": "a value",
            "additional_arg": None,
            "snakemake_args": ["--unknown-arg"],
            "snakemake_dir": Path("app").resolve(),
            "pybidsdb_dir": Path(db_path),
            "pybidsdb_reset": True,
            "pybids_db_dir": f"{DEPRECATION_FLAG}{db_path}{DEPRECATION_FLAG}",
            "pybids_db_reset": f"{DEPRECATION_FLAG}1{DEPRECATION_FLAG}",
            "snakefile": Path("app/Snakefile"),
            "bids_dir": Path("input_dataset").resolve(),
            "output_dir": Path(root).resolve(),
            "participant_label": None,
            "exclude_participant_label": None,
            "derivatives": [],
            "snakemake_version": "version",
            "snakebids_version": "version",
            "app_version": "version",
            "plugins.version.version": "version",
            "snakemake_target": "all",
            "analysis_level": "participant",
        }
    )
    if root == "app":
        expected_config["output_dir"] /= "results"

    app.run(
        args=[
            "input_dataset",
            root,
            "participant",
            "--derivatives",
            "--pybidsdb-dir",
            db_path,
            "--pybidsdb-reset",
            "--unknown-arg",
        ],
    )

    # First condition: outputdir is equal to snakemake_dir, or under
    # snakemake_dir/results
    if root in {"app", "app/results"}:
        cwd = Path("app").resolve()
        new_config = cwd / "config/config.yaml"
        expected_config["root"] = Path("results")
        io_mocks["write_output_mode"].assert_called_once_with(
            cwd / ".snakebids", "workflow"
        )
        io_mocks["write_config"].assert_called_once_with(
            config_file=new_config,
            data=expected_config,
            force_overwrite=True,
        )

    # Second condition: outputdir is an arbitrary path
    else:
        cwd = Path(root).resolve()
        expected_config["root"] = ""
        new_config = cwd / "config/config.yaml"
        io_mocks["write_output_mode"].assert_not_called()
        io_mocks["prepare_output"].assert_called_once_with(cwd, False)
        io_mocks["write_config"].assert_called_once_with(
            config_file=new_config,
            data=expected_config,
            force_overwrite=True,
        )

    io_mocks["snakemake"].assert_called_once_with(
        [
            "all",
            "--snakefile",
            str(plugin.snakefile_path),
            "--directory",
            str(cwd),
            "--configfile",
            str(new_config),
            "--unknown-arg",
        ]
    )



from __future__ import annotations

import itertools as it
import os
from collections.abc import Iterable
from pathlib import Path

import more_itertools as itx
import pytest
from hypothesis import assume, example, given
from hypothesis import strategies as st
from pathvalidate import Platform, is_valid_filename, is_valid_filepath

from snakebids.paths import bids, specs
from snakebids.paths._factory import bids_factory
from snakebids.paths._utils import BidsPathSpec
from snakebids.utils.utils import BidsEntity
from tests import strategies as sb_st
from tests.helpers import Benchmark, is_strictly_increasing


def _get_entity_tags(entities: Iterable[str]):
    return [BidsEntity.normalize(entity).tag for entity in entities]


def _values() -> st.SearchStrategy[str]:
    return st.text(min_size=1).filter(
        lambda s: is_valid_filename(s, Platform.LINUX) and "_" not in s and "-" not in s
    )


def _roots():
    return st.text().filter(lambda s: is_valid_filepath(s, Platform.LINUX))


def make_bids_testsuite(spec: BidsPathSpec):
    std_entities = {e["entity"] for e in spec}
    has_tag = {e["entity"] for e in spec if e.get("tag")}
    has_dir = {e["entity"] for e in spec if e.get("dir")}
    bids = bids_factory(spec)

    def _bids_args(
        entities: set[str] | None = std_entities,
        nonstandard: bool = True,
        custom: bool = True,
    ):
        std_ents = (
            sb_st.bids_entity(whitelist_entities=entities)
            if entities is not None
            else sb_st.nothing()
        )
        custom_ents = (
            _values()
            .map(BidsEntity)
            .filter(
                lambda s: str(s)
                not in (
                    {"datatype", "suffix", "extension", "prefix"} | set(std_entities)
                )
            )
            if custom
            else sb_st.nothing()
        )
        nonstd_ents = (
            sb_st.bids_entity(whitelist_entities=["datatype", "suffix", "extension"])
            if nonstandard
            else sb_st.nothing()
        )
        return (
            st.dictionaries(
                keys=std_ents | custom_ents | nonstd_ents,
                # The boolean here is to decide whether to use the entity or the tag in
                # the BidsEntity generated above
                values=st.tuples(
                    st.booleans(),
                    _values(),
                ),
                min_size=1,
            )
            .map(
                # we only want to use the full entity name if the spec in use
                # understands the full entity name
                lambda d: {
                    key.entity if val[0] and str(key) in has_tag else key.tag: val[1]
                    for key, val in d.items()
                }
            )
            .filter(lambda s: set(s) - {"datatype", "extension"})
        )

    class BidsTests:
        @given(_bids_args(nonstandard=False))
        def test_number_of_underscore_corresponds_to_number_entities(
            self, entities: dict[str, str]
        ):
            assert bids(**entities).count("_") == len(entities) - 1

        @given(_bids_args(nonstandard=False))
        def test_number_of_dashes_corresponds_to_number_entities(
            self, entities: dict[str, str]
        ):
            assert Path(bids(**entities)).name.count("-") == len(entities)

        @given(entities=_bids_args(), prefix=_values())
        def test_beginning_of_name_always_prefix(
            self, entities: dict[str, str], prefix: str
        ):
            assert Path(bids(prefix=prefix, **entities)).name.startswith(prefix)

        @given(
            entities=_bids_args(nonstandard=False),
            suffix=_values(),
            extension=_values(),
        )
        def test_end_of_path_always_suffix_extension(
            self, entities: dict[str, str], suffix: str, extension: str
        ):
            assert bids(suffix=suffix, extension=extension, **entities).endswith(
                suffix + extension
            )

        @given(entities=_bids_args(nonstandard=False), suffix=_values())
        def test_underscore_precedes_suffix(
            self, entities: dict[str, str], suffix: str
        ):
            assert bids(suffix=suffix, **entities)[len(suffix) * -1 - 1] == "_"

        @given(
            entities=_bids_args(nonstandard=False),
            suffix=_values() | st.none(),
            extension=_values(),
        )
        def test_underscore_does_not_precede_extension(
            self, entities: dict[str, str], suffix: str | None, extension: str
        ):
            assert (
                bids(suffix=suffix, extension=extension, **entities)[
                    len(extension) * -1 - 1
                ]
                != "_"
            )

        @given(entities=_bids_args(nonstandard=False))
        def test_no_underscore_at_end_if_no_suffix(self, entities: dict[str, str]):
            assert bids(**entities)[-1] != "_"

        @given(entities=_bids_args(nonstandard=True, entities=None))
        def test_no_underscore_at_beginning_if_only_suffix(
            self, entities: dict[str, str]
        ):
            assert bids(**entities)[0] != "_"

        @given(entities=_bids_args(), root=_roots())
        def test_beginning_of_path_always_root(
            self, entities: dict[str, str], root: str
        ):
            path = bids(root=root, **entities)
            assert path.startswith(root)
            length = len(root) - 1 if root[-1] == os.path.sep else len(root)
            assert path[length] == os.path.sep

        @example(entities={"sub": "0"}, datatype=".", root="0")
        @given(
            entities=_bids_args(nonstandard=False), datatype=_values(), root=_roots()
        )
        def test_bottom_directory_always_datatype(
            self, entities: dict[str, str], datatype: str, root: str
        ):
            # use os.path functions so that datatype=="." is treated safely
            assert datatype == os.path.basename(
                os.path.dirname(bids(root=root, datatype=datatype, **entities))
            )

        @given(entities=_bids_args(nonstandard=False), datatype=_values())
        def test_datatype_not_in_path_name(
            self, entities: dict[str, str], datatype: str
        ):
            assume(datatype not in "".join(it.chain.from_iterable(entities.items())))
            assert datatype not in Path(bids(datatype=datatype, **entities)).name

        @given(entities=_bids_args(nonstandard=False))
        def test_entities_all_in_path_as_tags(self, entities: dict[str, str]):
            tags = _get_entity_tags(entities)
            path = "_" + Path(bids(**entities)).name
            for tag in tags:
                assert f"_{tag}-" in path

        @given(entities=_bids_args(entities=has_tag, nonstandard=False, custom=False))
        def test_full_entity_names_not_in_path(self, entities: dict[str, str]):
            non_tags = [
                normed.entity
                for entity in entities
                if entity != (normed := BidsEntity.normalize(entity)).tag
            ]
            path = bids(**entities)
            for entity in non_tags:
                assert f"{entity}-" not in path

        @given(entities=_bids_args(entities=has_tag, nonstandard=False, custom=False))
        def test_long_and_short_names_cannot_be_used_simultaneously(
            self, entities: dict[str, str]
        ):
            entities = {BidsEntity.normalize(e).entity: v for e, v in entities.items()}
            tags = {BidsEntity.normalize(e).tag: v for e, v in entities.items()}
            with pytest.raises(
                ValueError,
                match="Long and short names of an entity cannot be used in the same",
            ) as err:
                bids(**entities, **tags)
            assert itx.first(entities) in err.value.args[0]
            assert itx.first(tags) in err.value.args[0]

        @given(
            entities=_bids_args(nonstandard=False, custom=False),
            custom=_bids_args(entities=None, nonstandard=False),
        )
        def test_entities_found_in_name_in_correct_order(
            self, entities: dict[str, str], custom: dict[str, str]
        ):
            # Make sure custom tags don't overlap with main tags
            assume(not set(entities) & set(custom))
            tags = _get_entity_tags(entities)
            order: list[str] = []
            for entity in spec:
                if (tag := entity.get("tag", entity["entity"])) in tags:
                    order.append(tag)
            # Custom tags come after defined tags
            order.extend(custom)
            path = "_" + Path(bids(**custom, **entities)).name
            assert set(tags) | set(custom) == set(order)
            assert is_strictly_increasing(path.index(f"_{e}-") for e in order)

        @given(
            entities=_bids_args(entities=std_entities - has_dir, nonstandard=False),
            root=_roots(),
        )
        def test_nondir_entities_dont_have_dirs(
            self, entities: dict[str, str], root: str
        ):
            assert Path(bids(root=root, **entities)).parent == Path(root)

        @given(entities=_bids_args(entities=has_dir, nonstandard=False, custom=False))
        def test_dir_entities_each_own_dir(self, entities: dict[str, str]):
            for par in itx.islice_extended(Path(bids(**entities)).parents, 0, -1):
                count = 0
                for e in list(entities):
                    tag = BidsEntity.normalize(e).tag
                    if (
                        par.name[: len(tag)] == tag
                        # if found, appears nowhere else
                        and f"{tag}-" not in str(par.parent)
                    ):
                        del entities[e]
                        count += 1
                assert count == 1
            # Check that all entities have been removed (ie found)
            assert not entities

        @given(
            entities=_bids_args(entities=has_dir, nonstandard=False, custom=False),
            root=_roots().filter(lambda s: s != "."),
        )
        def test_directories_in_correct_order(
            self, entities: dict[str, str], root: str
        ):
            tags = _get_entity_tags(entities)
            order: list[str] = []
            for entity in spec:
                if (tag := entity.get("tag", entity["entity"])) in tags:
                    order.append(tag)
            path = str(Path(bids(root=root, **entities)).parent)
            assert set(tags) == set(order)
            assert is_strictly_increasing(
                path.index(f"{os.path.sep}{e}-") for e in order
            )

        @given(entities=_bids_args(nonstandard=False), root=_roots())
        def test_values_paired_with_entities(self, entities: dict[str, str], root: str):
            path = Path(bids(root=root, **entities))
            name = "_" + path.name
            parent = os.path.sep + str(path.parent)

            def assert_follows(string: str, first: str, second: str):
                start = string.index(first) + len(first)
                assert string[start : start + len(second)] == second

            for entity, value in entities.items():
                tag = BidsEntity.normalize(entity).tag
                assert_follows(name, f"_{tag}-", value)
                if entity in has_dir:
                    assert_follows(parent, f"{os.path.sep}{tag}-", value)

        def test_bids_with_no_args_gives_empty_path(self):
            assert not bids()

        @given(
            args=st.dictionaries(
                st.sampled_from(["datatype", "prefix"]), _values(), min_size=1
            )
        )
        def test_missing_essential_entities_gives_error(self, args: dict[str, str]):
            with pytest.raises(
                ValueError,
                match="At least one of suffix, extension, or an entity must be",
            ):
                bids(**args)

    return BidsTests


TestV0_0_0 = make_bids_testsuite(specs.v0_0_0())

TestV0_10_1 = make_bids_testsuite(specs.v0_11_0())


def test_benchmark_bids(benchmark: Benchmark):
    """If refactoring bids, be sure this benchmark doesn't needlessly increase"""
    benchmark(
        bids,
        root="foo/bar",
        subject_dir=False,
        subject="001",
        session="32",
        run="stop",
        foo="bar",
        england="britain",
        space="cosmos",
        rome="fell",
        suffix="suffix",
        extension=".ext",
    )



from __future__ import annotations

import jinja2
from colorama import Fore
from jinja2.ext import Extension


class ColoramaExtension(Extension):
    """Include colorama foreground colors in the global environment."""

    def __init__(self, env: jinja2.Environment):
        super().__init__(env)
        env.globals["Fore"] = Fore  # type: ignore



from __future__ import annotations

import argparse
import sys
from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias, TypedDict

import attrs
import pluggy
from attrs.converters import default_if_none
from typing_extensions import Unpack

from snakebids.bidsapp import hookspecs
from snakebids.bidsapp.args import ArgumentGroups

if TYPE_CHECKING:
    from argparse import _ArgumentGroup, _FormatterClass

    from pluggy._hooks import _Plugin


class ArgumentParserArgs(TypedDict, total=False):
    """Arguments passed on to :class:`argparse.ArgumentParser`."""

    prog: str | None
    usage: str | None
    description: str | None
    epilog: str | None
    parents: Sequence[argparse.ArgumentParser]
    formatter_class: _FormatterClass
    prefix_chars: str
    fromfile_prefix_chars: str | None
    argument_default: Any
    conflict_handler: str
    add_help: bool
    allow_abbrev: bool
    exit_on_error: bool


def app(
    plugins: Iterable[_Plugin] | None = None,
    *,
    config: dict[str, Any] | None = None,
    **argparse_args: Unpack[ArgumentParserArgs],
) -> _Runner:
    """Create a BIDSApp.

    Parameters
    ----------
    plugins
        List of snakebids plugins to apply to app (see :ref:`using-plugins`).
    config
        Initial config. Will be updated with parsed arguments from ``parser`` and
        potentially modified by plugins
    **argparse_args
        Arguments passed on transparently to :class:`argparse.ArgumentParser`.
    """
    pm = pluggy.PluginManager("snakebids")
    pm.add_hookspecs(hookspecs)
    parser = argparse.ArgumentParser(**argparse_args)
    if plugins is not None:
        _add_plugin_dependencies(pm, plugins, registered_ok=False)
    return _Runner(pm=pm, parser=parser, config=config)


class _DependingPlugin(Protocol):
    DEPENDENCIES: tuple[_Plugin, ...]


def _add_plugin_dependencies(
    pm: pluggy.PluginManager,
    plugins: Iterable[_Plugin | _DependingPlugin],
    *,
    registered_ok: bool,
) -> None:
    dependencies: list[_Plugin] = []
    for plugin in plugins:
        if registered_ok and pm.is_registered(plugin):
            continue
        pm.register(plugin)
        if hasattr(plugin, "DEPENDENCIES"):
            dependencies.extend(plugin.DEPENDENCIES)  # type: ignore
    if not dependencies:
        return None
    return _add_plugin_dependencies(pm, dependencies, registered_ok=True)


if TYPE_CHECKING:
    _ArgumentGroupDictType: TypeAlias = "defaultdict[str, _ArgumentGroup]"
else:
    _ArgumentGroupDictType = defaultdict


class _ArgumentGroupDict(_ArgumentGroupDictType):
    """Dict that converts missing keys into argument groups."""

    def __init__(self, parser: argparse.ArgumentParser):
        self.parser = parser

    def __missing__(self, key: str):
        new = self.parser.add_argument_group(key)
        self[key] = new
        return new


@attrs.define
class _Runner:
    """Runtime manager for BIDS apps.

    Manages the parser, config, and plugin relay for snakebids BIDS apps. This class
    should not be constructed directly, but built using the
    :func:`bidsapp.app() <snakebids.bidsapp.app>` function.
    """

    pm: pluggy.PluginManager
    """Reference to the plugin manager."""

    parser: argparse.ArgumentParser
    """Parser used for parsing CLI arguments.

    The parser may be manipulated before running action methods to add initial arguments
    (via :meth:`parser.add_argument() <argparse.ArgumentParser.add_argument>`) and
    argument groups (via :meth:`parser.add_argument_group()
    <argparse.ArgumentParser.add_argument_group>`). Any argument groups added will
    automatically be indexed in :attr:`.argument_groups` and made available to plugins
    via their ``title``.
    """

    config: dict[str, Any] = attrs.field(converter=default_if_none(factory=dict))
    """Configuration dictionary for passing data between plugins."""

    argument_groups: ArgumentGroups = attrs.field(
        default=attrs.Factory(
            lambda self: _ArgumentGroupDict(self.parser), takes_self=True
        )
    )
    """Argument group reference accessible to plugins for organizing CLI arguments.

    Any groups added to the :attr:`parser` before the action methods are called will
    be automatically added, indexed by their titles. Thus, this object should typically
    only be used within plugins.
    """

    _parser_built: bool = attrs.field(default=False, init=False)
    _processed: bool = attrs.field(default=False, init=False)
    _run: bool = attrs.field(default=False, init=False)

    def build_parser(self):
        """Run plugins affecting the ``parser`` without yet parsing arguments."""
        if not self._parser_built:
            self.pm.hook.initialize_config(config=self.config)
            for group in self.parser._action_groups:  # noqa: SLF001
                if group.title is not None:
                    self.argument_groups[group.title] = group
            self.pm.hook.add_cli_arguments(
                parser=self.parser,
                argument_groups=self.argument_groups,
                config=self.config,
            )
            self._parser_built = True
        return self

    def parse_args(self, args: list[str] | None = None):
        """Run all plugins and parse arguments."""
        if not self._processed:
            self.build_parser()
            args = sys.argv[1:] if args is None else args
            argv: list[str] | None = self.pm.hook.get_argv(
                argv=args, config=self.config
            )
            namespace, unknown = self.parser.parse_known_args(
                args=args if argv is None else argv
            )
            self.pm.hook.handle_unknown_args(args=unknown, config=self.config)
            self.pm.hook.update_cli_namespace(
                namespace=namespace.__dict__, config=self.config
            )
            self.config.update(namespace.__dict__)
            self.pm.hook.finalize_config(config=self.config)
            self._processed = True
        return self

    def run(self, args: list[str] | None = None):
        if not self._run:
            self.parse_args(args)
            self.pm.hook.run(config=self.config)
            self._run = True



from __future__ import annotations

import argparse
from typing import Any

import pluggy

from snakebids.bidsapp.args import ArgumentGroups

hookspec = pluggy.HookspecMarker("snakebids")


@hookspec
def initialize_config(config: dict[str, Any]):
    """Modify config dict before creation of CLI parser.

    Config should be modified in place.

    Parameters
    ----------
    config
        Possibly empty dictionary of configuration values
    """


@hookspec
def add_cli_arguments(
    parser: argparse.ArgumentParser,
    config: dict[str, Any],
    argument_groups: ArgumentGroups,
):
    """Add any number of arguments to the argument parser.

    If special intervention is not made in the :func:`update_cli_namespace` hook,
    argument data will be automatically merged into ``config`` following argument
    parsing under the argument `dest`. To avoid naming conflicts, plugins should
    explicitly set the `dest` of each argument they add to
    ``plugins.<plugin_name>.<argument_name>``.

    Parameters
    ----------
    parser
        BIDS app CLI parser
    config
        Configuration dictionary preloaded with values from :func:`initialize_config`
    """


@hookspec(firstresult=True)
def get_argv(argv: list[str], config: dict[str, Any]) -> list[str] | None:
    """Set or modify the CLI parameters that will be parsed by the parser."""


@hookspec
def handle_unknown_args(args: list[str], config: dict[str, Any]):
    """If ``parse_known_args`` enabled, handle unknown arguments.

    Parameters
    ----------
    args
        List of unknown arguments parsed by the argparse parser
    config
        Configuration dictionary
    """


@hookspec
def update_cli_namespace(namespace: dict[str, Any], config: dict[str, Any]):
    """Interact with argument parsing results before they are merged into ``config``.

    The ``namespace`` contains the results from
    :meth:`~argparse.ArgumentParser.parse_args` (equivalent to
    :class:`vars(Namespace) <argparse.Namespace>`). Any modifications made to this
    :class:`dict` will be carried forward in app initialization. For instance, if
    an entry is deleted from ``namespace``, it will not be available to downstream
    plugins or be copied into ``config``.

    Parameters
    ----------
    namespace
        the dictionary of values derived from the argparse namespace
    config
        Configuration dictionary
    """


@hookspec
def finalize_config(config: dict[str, Any]):
    """Perform modifications to the config following merging of parsed arguments.

    Parameters
    ----------
    config
        Configuration dictionary
    """


@hookspec
def run(config: dict[str, Any]):
    """Consume configuration and perform actions.

    This hook runs directly after :func:`finalize_config`. The config should not be
    modified.

    Parameters
    ----------
    config
        The finalized configuration dictionary
    """
